import itertools
import sys
import typing as tp
from copy import deepcopy
from functools import reduce

import numpy as np
from arraykit import array_deepcopy
from arraykit import first_true_1d
from automap import FrozenAutoMap  # pylint: disable = E0611
from automap import NonUniqueError  # pylint: disable=E0611

from static_frame.core.exception import ErrorInitIndexNonUnique
from static_frame.core.exception import LocEmpty
from static_frame.core.exception import LocInvalid
from static_frame.core.util import DTYPE_BOOL
from static_frame.core.util import DTYPE_DATETIME_KIND
from static_frame.core.util import DTYPE_OBJECT
from static_frame.core.util import DTYPE_OBJECTABLE_DT64_UNITS
from static_frame.core.util import DTYPE_UINT_DEFAULT
from static_frame.core.util import EMPTY_ARRAY_INT
from static_frame.core.util import EMPTY_FROZEN_AUTOMAP
from static_frame.core.util import EMPTY_SLICE
from static_frame.core.util import INT_TYPES
from static_frame.core.util import NULL_SLICE
from static_frame.core.util import OPERATORS
from static_frame.core.util import SLICE_ATTRS
from static_frame.core.util import SLICE_START_ATTR
from static_frame.core.util import SLICE_STEP_ATTR
from static_frame.core.util import SLICE_STOP_ATTR
from static_frame.core.util import GetItemKeyType

if tp.TYPE_CHECKING:
    from static_frame.core.index import Index  # pylint: disable=W0611,C0412 # pragma: no cover


HierarchicalLocMapKey = tp.Union[np.ndarray, tp.Tuple[tp.Union[tp.Sequence[tp.Hashable], tp.Hashable], ...]]
_HLMap = tp.TypeVar('_HLMap', bound='HierarchicalLocMap')
TypePos = tp.Optional[int]
LocEmptyInstance = LocEmpty()


class FirstDuplicatePosition(KeyError):
    def __init__(self, first_dup: int) -> None:
        self.first_dup = first_dup


class LocMap:

    @staticmethod
    def map_slice_args(
            label_to_pos: tp.Callable[[tp.Iterable[tp.Hashable]], int],
            key: slice,
            labels: tp.Optional[np.ndarray] = None,
            ) -> tp.Iterator[tp.Union[int, None]]:
        '''Given a slice ``key`` and a label-to-position mapping, yield each integer argument necessary to create a new iloc slice. If the ``key`` defines a region with no constituents, raise ``LocEmpty``

        Args:
            label_to_pos: callable into mapping (can be a get() method from a dictionary)
        '''
        # NOTE: it is expected that NULL_SLICE is already identified
        labels_astype: tp.Optional[np.ndarray] = None

        for field in SLICE_ATTRS:
            attr = getattr(key, field)
            if attr is None:
                yield None

            # NOTE: We can do `is` checks on field since `SLICE_ATTRS` only contains some certain symbolic constants

            elif attr.__class__ is np.datetime64:
                if field is SLICE_STEP_ATTR:
                    raise RuntimeError(f'Step cannot be {attr}')
                # if we match the same dt64 unit, simply use label_to_pos, increment stop
                if attr.dtype == labels.dtype: # type: ignore
                    pos: TypePos = label_to_pos(attr)
                    if pos is None:
                        # if same type, and that atter is not in labels, we fail, just as we do in then non-datetime64 case. Only when datetimes are given in a different unit are we "loose" about matching.
                        raise LocInvalid('Invalid loc given in a slice', attr, field)
                    if field is SLICE_STOP_ATTR:
                        pos += 1 # stop is inclusive
                elif field is SLICE_START_ATTR:
                    # NOTE: as an optimization only for the start attr, we can try to convert attr to labels unit and see if there is a match; this avoids astyping the entire labels array
                    pos: TypePos = label_to_pos(attr.astype(labels.dtype)) #type: ignore
                    if pos is None: # we did not find a start position
                        labels_astype = labels.astype(attr.dtype) #type: ignore
                        matches = np.flatnonzero(labels_astype == attr)
                        if len(matches):
                            pos = matches[0]
                        else:
                            raise LocEmptyInstance
                elif field is SLICE_STOP_ATTR:
                    # NOTE: we do not want to convert attr to labels dtype and take the match as we want to get the last of all possible matches of labels at the attr unit
                    # NOTE: try to re-use labels_astype if possible
                    if labels_astype is None or labels_astype.dtype != attr.dtype:
                        labels_astype = labels.astype(attr.dtype) #type: ignore
                    matches = np.flatnonzero(labels_astype == attr)
                    if len(matches):
                        pos = matches[-1] + 1
                    else:
                        raise LocEmptyInstance

                yield pos

            else:
                if field is not SLICE_STEP_ATTR:
                    pos = label_to_pos(attr)
                    if pos is None:
                        # NOTE: could raise LocEmpty() to silently handle this
                        raise LocInvalid('Invalid loc given in a slice', attr, field)
                else: # step
                    pos = attr # should be an integer
                    if not isinstance(pos, INT_TYPES):
                        raise TypeError(f'Step must be an integer, not {pos}')
                if field is SLICE_STOP_ATTR:
                    # loc selections are inclusive, so iloc gets one more
                    pos += 1
                yield pos

    @classmethod
    def loc_to_iloc(cls, *,
            label_to_pos: tp.Dict[tp.Hashable, int],
            labels: np.ndarray,
            positions: np.ndarray,
            key: GetItemKeyType,
            partial_selection: bool = False,
            ) -> GetItemKeyType:
        '''
        Note: all SF objects (Series, Index) need to be converted to basic types before being passed as `key` to this function.

        Args:
            partial_selection: if True and key is an iterable of labels that includes labels not in the mapping, available matches will be returned rather than raising.
        Returns:
            An integer mapped slice, or GetItemKey type that is based on integers, compatible with TypeBlocks
        '''
        # NOTE: ILoc is handled prior to this call, in the Index._loc_to_iloc method

        if key.__class__ is slice:
            if key == NULL_SLICE:
                return NULL_SLICE
            try:
                return slice(*cls.map_slice_args(
                        label_to_pos.get, #type: ignore
                        key,
                        labels)
                        )
            except LocEmpty:
                return EMPTY_SLICE

        labels_is_dt64 = labels.dtype.kind == DTYPE_DATETIME_KIND

        if key.__class__ is np.datetime64:
            # if we have a single dt64, convert this to the key's unit and do a Boolean selection if the key is a less-granular unit
            if (labels.dtype == DTYPE_OBJECT
                    and np.datetime_data(key.dtype)[0] in DTYPE_OBJECTABLE_DT64_UNITS): #type: ignore
                key = key.astype(DTYPE_OBJECT) #type: ignore
            elif labels_is_dt64 and key.dtype < labels.dtype: #type: ignore
                key = labels.astype(key.dtype) == key #type: ignore
            # if not different type, keep it the same so as to do a direct, single element selection

        is_array = key.__class__ is np.ndarray
        is_list = isinstance(key, list)

        # can be an iterable of labels (keys) or an iterable of Booleans
        if is_array or is_list:
            if len(key) == 0:
                return EMPTY_ARRAY_INT

            if is_array and key.dtype.kind == DTYPE_DATETIME_KIND: #type: ignore
                dt64_unit = np.datetime_data(key.dtype)[0] #type: ignore
                # NOTE: only in the conditions of an empty array, the unit might be generic
                if (labels.dtype == DTYPE_OBJECT and dt64_unit in DTYPE_OBJECTABLE_DT64_UNITS):
                    # if key is dt64 and labels are object, then for objectable units we can convert key to object to permit matching in the AutoMap
                    # NOTE: tolist() is expected to be faster than astype object for smaller collections
                    key = key.tolist() #type: ignore
                    is_array = False
                    is_list = True
                elif labels_is_dt64 and key.dtype < labels.dtype: #type: ignore
                    # NOTE: change the labels to the dt64 dtype, i.e., if the key is years, recast the labels as years, and do a Boolean selection of everything that matches each key
                    labels_ref = labels.astype(key.dtype) # type: ignore
                    # NOTE: this is only correct if both key and labels are dt64, and key is a less granular unit, as the order in the key and will not be used
                    # let Boolean key advance to next branch
                    key = reduce(OPERATORS['__or__'], (labels_ref == k for k in key)) # type: ignore

            if is_array and key.dtype == DTYPE_BOOL: #type: ignore
                return positions[key]

            # map labels to integer positions, return a list of integer positions
            # NOTE: we may miss the opportunity to identify contiguous keys and extract a slice
            # NOTE: we do more branching here to optimize performance
            if partial_selection:
                return [label_to_pos[k] for k in key if k in label_to_pos] # type: ignore
            return [label_to_pos[k] for k in key] # type: ignore

        # if a single element (an integer, string, or date, we just get the integer out of the map
        return label_to_pos[key] #type: ignore


class HierarchicalLocMap:
    '''
    A utility utilized by IndexHierarchy in order to quickly map keys to ilocs.
    '''

    __slots__ = (
            'bit_offset_encoders',
            'encoding_can_overflow',
            'encoded_indexer_map',
            )

    bit_offset_encoders: np.ndarray
    encoding_can_overflow: bool
    encoded_indexer_map: FrozenAutoMap

    def __init__(self: _HLMap,
            *,
            indices: tp.List['Index'],
            indexers: tp.List[np.ndarray],
            ) -> None:

        if not len(indexers[0]):
            self.bit_offset_encoders = np.full(len(indices), 0, dtype=DTYPE_UINT_DEFAULT)
            self.encoding_can_overflow = False
            self.encoded_indexer_map = EMPTY_FROZEN_AUTOMAP
            return

        self.bit_offset_encoders, self.encoding_can_overflow = self.build_offsets_and_overflow(
                num_unique_elements_per_depth=list(map(len, indices))
                )
        try:
            self.encoded_indexer_map = self.build_encoded_indexers_map(
                    encoding_can_overflow=self.encoding_can_overflow,
                    bit_offset_encoders=self.bit_offset_encoders,
                    indexers=indexers,
                    )
        except FirstDuplicatePosition as e:
            duplicate_labels = tuple(
                    index[indexer[e.first_dup]]
                    for (index, indexer) in zip(indices, indexers)
                    )
            raise ErrorInitIndexNonUnique(duplicate_labels) from None

    def __deepcopy__(self: _HLMap,
            memo: tp.Dict[int, tp.Any],
            ) -> _HLMap:
        '''
        Return a deep copy of this IndexHierarchy.
        '''
        obj: _HLMap = self.__class__.__new__(self.__class__)
        obj.bit_offset_encoders = array_deepcopy(self.bit_offset_encoders, memo)
        obj.encoding_can_overflow = self.encoding_can_overflow
        obj.encoded_indexer_map = deepcopy(self.encoded_indexer_map, memo)

        memo[id(self)] = obj
        return obj

    def __setstate__(self, state: tp.Tuple[None, tp.Dict[str, tp.Any]]) -> None:
        '''
        Ensure that reanimated NP arrays are set not writeable.
        '''
        for key, value in state[1].items():
            setattr(self, key, value)
        self.bit_offset_encoders.flags.writeable = False

    @property
    def nbytes(self: _HLMap) -> int:
        return (
                sys.getsizeof(self.encoding_can_overflow) +
                tp.cast(int, self.bit_offset_encoders.nbytes) +
                sys.getsizeof(self.encoded_indexer_map)
        )

    @staticmethod
    def build_offsets_and_overflow(
            num_unique_elements_per_depth: tp.List[int],
            ) -> tp.Tuple[np.ndarray, bool]:
        '''
        Derive the offsets and the overflow flag from the number of unique values per depth
        '''
        # `bit_sizes` is an array that shows how many bits are needed to contain the max indexer per depth
        #
        # For example, lets say there are 3 levels, and number of unique elements per depth is [71, 5, 13].
        # `bit_sizes` will become: [7, 3, 4], which mean that we can fit the indexer for each depth into 7, 3, and 4 bits, respectively.
        #    (i.e.) all valid indexers at depth N can be represented with bit_sizes[N] bits!
        #
        # We see this: [2**power for power in [7,3,4]] = [128, 8, 16]
        # We can prove this: num_unique_elements_per_depth <= [2**bit_size for bit_size in bit_sizes]
        bit_sizes = np.floor(np.log2(num_unique_elements_per_depth)) + 1

        # Based on bit_sizes, we cumsum to determine the successive number of total bits needed for each depth
        # Using the previous example, this would look like: [7, 10, 14]
        # This means:
        #  - depth 0 ends at bit offset 7.
        #  - depth 1 ends at bit offset 10. (depth 1 needs 3 bits!)
        #  - depth 2 ends at bit offset 14. (depth 2 needs 4 bits!)
        bit_end_positions = np.cumsum(bit_sizes, dtype=DTYPE_UINT_DEFAULT)

        # However, since we ultimately need these values to bitshift, we want them to offset based on start position, not end.
        # This means:
        #  - depth 0 starts at bit offset 0.
        #  - depth 1 starts at bit offset 7. (depth 0 needed 7 bits!)
        #  - depth 2 starts at bit offset 10. (depth 1 needed 3 bits!)
        bit_start_positions = np.zeros(
                len(bit_end_positions),
                dtype=DTYPE_UINT_DEFAULT)
        bit_start_positions[1:] = bit_end_positions[:-1]
        bit_start_positions.flags.writeable = False

        # We now return these offsets, and whether or not we have overflow.
        # If the last end bit is greater than 64, then it means we cannot encode a label's indexer into a uint64.
        return bit_start_positions, bit_end_positions[-1] > 64

    @staticmethod
    def build_encoded_indexers_map(
            *,
            encoding_can_overflow: bool,
            bit_offset_encoders: np.ndarray,
            indexers: np.ndarray,
            ) -> FrozenAutoMap:
        '''
        Builds up a mapping from indexers to iloc positions using their encoded values
        '''
        # We previously determined we cannot encode indexers into uint64. Cast to object to rely on Python's bigint
        if encoding_can_overflow:
            indexers = indexers.astype(object).T
        else:
            indexers = indexers.astype(DTYPE_UINT_DEFAULT).T

        # Encode indexers into uint64
        # indexers: (n, m), offsets: (m,)
        # This bitshifts all indexers by the offset, resulting in numbers that are ready to be bitwise OR'd
        # We need to reverse in order to have depth 0
        # Example:
        #  indexers:      bitshift         (Bit representation)
        #                                    d0, d1 d0, d2 d1 d0  (d = depth)
        #    [0, 1, 2] => [0, 4, 32]       ([00, 01 00, 10 00 00])
        #    [0, 2, 0] => [0, 8,  0]       ([00, 10 00, 00 00 00])
        #    [2, 2, 0] => [2, 8,  0]       ([10, 10 00, 00 00 00])
        #    [1, 0, 1] => [1, 0, 16]       ([01, 00 00, 01 00 00])
        encoded_indexers = indexers << bit_offset_encoders

        # Finally, we bitwise OR all them together to encode them into a single, unique uint64 for each iloc
        #  encoded_indexers   bitwise OR   (Bit representation)
        #                                    d0 d1 d2  (d = depth)
        #    [2, 4,  0]    => [36]         ([10 01 00])
        #    [0, 8,  0]    => [ 8]         ([00 10 00])
        #    [0, 8, 32]    => [10]         ([00 10 10])
        #    [1, 0, 16]    => [17]         ([01 00 01])
        encoded_indexers = np.bitwise_or.reduce(encoded_indexers, axis=1)

        # Success! We have now successfully encoded each indexer into a single, unique uint64.
        #    [0, 1, 2] => [36]
        #    [0, 2, 0] => [ 8]
        #    [2, 2, 0] => [10]
        #    [1, 0, 1] => [17]

        # Finally, we create a mapping from encoded indexers to ilocs.
        #    [0, 1, 2] => [36] => [0]
        #    [0, 2, 0] => [ 8] => [1]
        #    [2, 2, 0] => [10] => [2]
        #    [1, 0, 1] => [17] => [3]
        # len(encoded_indexers) == len(self)!
        try:
            return FrozenAutoMap(encoded_indexers.tolist()) # Automap is faster with Python lists :(
        except NonUniqueError as e:
            # nonzero returns arrays of indices per dimension. We are 1D, so we
            # will receive an array containing one other array. Of that inner
            # array, we only need the first occurrence
            first_duplicate = first_true_1d(encoded_indexers == e.args[0], forward=True)
            raise FirstDuplicatePosition(first_duplicate) from None

    @staticmethod
    def is_single_element(element: tp.Any) -> bool:
        # By definition, all index labels are hashable. If it's not, then it
        # means this must be a container of labels.
        try:
            hash(element)
        except TypeError:
            return False
        return True

    def build_key_indexers(self: _HLMap,
            key: HierarchicalLocMapKey,
            indices: tp.List['Index'],
            ) -> np.ndarray:
        key_indexers: tp.List[tp.Sequence[int]] = []

        is_single_key = True

        subkey_indexers: tp.List[int]

        # 1. Perform label resolution
        for key_at_depth, index_at_depth in zip(key, indices):
            if self.is_single_element(key_at_depth):
                key_indexers.append((index_at_depth._loc_to_iloc(key_at_depth),))
            else:
                is_single_key = False
                subkey_indexers = []
                for sub_key in key_at_depth:
                    subkey_indexers.append(index_at_depth._loc_to_iloc(sub_key))
                key_indexers.append(subkey_indexers)

        # 2. Convert to numpy array
        combinations = np.array(list(itertools.product(*key_indexers)), dtype=DTYPE_UINT_DEFAULT)
        if is_single_key and len(combinations) == 1:
            [combinations] = combinations

        if self.encoding_can_overflow:
            return combinations.astype(object)

        return combinations

    def loc_to_iloc(self: _HLMap,
            key: HierarchicalLocMapKey,
            indices: tp.List['Index'],
            ) -> tp.Union[int, tp.List[int]]:
        key_indexers = self.build_key_indexers(key=key, indices=indices)

        # 2. Encode the indexers. See `build_encoded_indexers_map` for detailed comments.
        key_indexers <<= self.bit_offset_encoders

        if key_indexers.ndim == 2:
            key_indexers = np.bitwise_or.reduce(key_indexers, axis=1)
            return list(map(self.encoded_indexer_map.__getitem__, key_indexers))

        key_indexers = np.bitwise_or.reduce(key_indexers)
        return self.encoded_indexer_map[key_indexers] # type: ignore

    def indexers_to_iloc(self: _HLMap,
            indexers: np.ndarray,
            ) -> tp.List[int]:
        '''
        Encodes indexers, and then remaps them to ilocs using the encoded_indexer_map
        '''
        indexers = self.encode(indexers, self.bit_offset_encoders)
        return list(map(self.encoded_indexer_map.__getitem__, indexers))

    @staticmethod
    def encode(indexers: np.ndarray, bit_offset_encoders: np.ndarray) -> np.ndarray:
        '''
        Encode indexers into a 1-dim array of uint64
        '''
        # Validate input requirements
        assert indexers.ndim == 2
        assert indexers.shape[1] == len(bit_offset_encoders)
        assert indexers.dtype == DTYPE_UINT_DEFAULT

        return np.bitwise_or.reduce(indexers << bit_offset_encoders, axis=1)

    @staticmethod
    def unpack_encoding(
            encoded_arr: np.ndarray,
            bit_offset_encoders: np.ndarray,
            encoding_can_overflow: bool,
            ) -> np.ndarray:
        '''
        Given an encoding, unpack it into its constituent parts

        Ex:
            bit_offset_encoders = [0, 2, 4]

            Encodings:
                36  => [0, 4, 32] => [0, 1, 2]
                 8  => [0, 8,  0] => [0, 2, 0]
                10  => [2, 8,  0] => [2, 2, 0]
                17  => [1, 0, 16] => [1, 0, 1]

            Step 1:
            Expand bit_offset_encoders into something more helpful -> masks

            [0, 2, 4] is where the bit offsets start. They end one bit before the next offset.
            Thus, the bit offset ends are:
            [1, 3, 64] # since 64 is the max bit offset

            From here, we build up a list of masks that each have the correct number of up bits

            0 => [11] # 2 bit mask
            2 => [11] # 2 bit mask
            4 => [11] # 2 bit mask

            Now, for each component (i.e. the number of bit_offset_encoders), we
            apply to corresponding mask to the values AFTER they have been shifted backwards

            36 == [10 01 00]
             8 == [00 10 00]
            10 == [00 10 10]
            17 == [01 00 01]

            Depth 0:
                offset = bit_offset_encoders[0] = 0
                36 => ([10 01 00] << 0) => [10 01 00] & [11] => [00 00 00] => 0
                 8 => ([00 10 00] << 0) => [00 10 00] & [11] => [00 00 00] => 0
                10 => ([00 10 10] << 0) => [00 10 10] & [11] => [00 00 10] => 2
                17 => ([01 00 01] << 0) => [01 00 01] & [11] => [00 00 01] => 1

            Depth 1:
                offset = bit_offset_encoders[1] = 2
                36 => ([10 01 00] << 2) => [10 01] & [11] => [00 01] => 1
                 8 => ([00 10 00] << 2) => [00 10] & [11] => [00 10] => 2
                10 => ([00 10 10] << 2) => [00 10] & [11] => [00 10] => 2
                17 => ([01 00 01] << 2) => [01 00] & [11] => [00 00] => 0

            Depth 2:
                offset = bit_offset_encoders[2] = 4
                36 => ([10 01 00] << 4) => [10] & [11] => [10] => 2
                 8 => ([00 10 00] << 4) => [00] & [11] => [00] => 0
                10 => ([00 10 10] << 4) => [00] & [11] => [00] => 0
                17 => ([01 00 01] << 4) => [01] & [11] => [01] => 1

            Result:
                36 => [0, 1, 2]
                 8 => [0, 2, 0]
                10 => [2, 2, 0]
                17 => [1, 0, 1]

            NOTE: This is the inverse of the documentation in `build_encoded_indexers_map`
        '''
        assert bit_offset_encoders.dtype == DTYPE_UINT_DEFAULT
        assert bit_offset_encoders[0] == 0 # By definition, the first offset starts at 0!
        assert encoded_arr.ndim == 1 # Encodings are always 1D

        dtype = object if encoding_can_overflow else DTYPE_UINT_DEFAULT

        starts = bit_offset_encoders
        stops = np.empty(len(starts), dtype=dtype)
        stops[:-1] = starts[1:]
        stops[-1] = 64

        lens = stops - starts
        masks = [x for x in (1 << lens) - 1]

        target = np.empty((len(bit_offset_encoders), len(encoded_arr)), dtype=DTYPE_UINT_DEFAULT)

        for depth in range(len(bit_offset_encoders)):
            target[depth] = (encoded_arr >> starts[depth]) & masks[depth]

        target.flags.writeable = False
        return target



import typing as tp
from collections import Counter
from copy import deepcopy
from itertools import chain
from itertools import zip_longest

import numpy as np
from arraykit import array_deepcopy
from arraykit import immutable_filter
from arraykit import mloc
from arraykit import name_filter
from arraykit import resolve_dtype
from automap import AutoMap  # pylint: disable=E0611
from automap import FrozenAutoMap  # pylint: disable=E0611
from automap import NonUniqueError  # pylint: disable=E0611

from static_frame.core.container import ContainerOperand
from static_frame.core.container_util import apply_binary_operator
from static_frame.core.container_util import iter_component_signature_bytes
from static_frame.core.container_util import key_from_container_key
from static_frame.core.container_util import matmul
from static_frame.core.container_util import sort_index_for_order
from static_frame.core.display import Display
from static_frame.core.display import DisplayActive
from static_frame.core.display import DisplayHeader
from static_frame.core.display_config import DisplayConfig
from static_frame.core.doc_str import doc_inject
from static_frame.core.exception import ErrorInitIndex
from static_frame.core.exception import ErrorInitIndexNonUnique
from static_frame.core.index_base import IndexBase
from static_frame.core.loc_map import LocMap
from static_frame.core.node_dt import InterfaceDatetime
from static_frame.core.node_iter import IterNodeApplyType
from static_frame.core.node_iter import IterNodeDepthLevel
from static_frame.core.node_iter import IterNodeType
from static_frame.core.node_re import InterfaceRe
from static_frame.core.node_selector import InterfaceGetItem
from static_frame.core.node_selector import InterfaceSelectDuo
from static_frame.core.node_selector import TContainer
from static_frame.core.node_str import InterfaceString
from static_frame.core.node_values import InterfaceValues
from static_frame.core.style_config import StyleConfig
from static_frame.core.util import DEFAULT_SORT_KIND
from static_frame.core.util import DTYPE_BOOL
from static_frame.core.util import DTYPE_DATETIME_KIND
from static_frame.core.util import DTYPE_INT_DEFAULT
from static_frame.core.util import DTYPE_NA_KINDS
from static_frame.core.util import DTYPE_OBJECT
from static_frame.core.util import DTYPE_OBJECTABLE_KINDS
from static_frame.core.util import EMPTY_ARRAY
from static_frame.core.util import INT_TYPES
from static_frame.core.util import KEY_ITERABLE_TYPES
from static_frame.core.util import NAME_DEFAULT
from static_frame.core.util import NULL_SLICE
from static_frame.core.util import DepthLevelSpecifier
from static_frame.core.util import DtypeSpecifier
from static_frame.core.util import GetItemKeyType
from static_frame.core.util import IndexConstructor
from static_frame.core.util import IndexInitializer
from static_frame.core.util import KeyIterableTypes
from static_frame.core.util import KeyTransformType
from static_frame.core.util import NameType
from static_frame.core.util import PositionsAllocator
from static_frame.core.util import UFunc
from static_frame.core.util import argsort_array
from static_frame.core.util import array2d_to_tuples
from static_frame.core.util import array_sample
from static_frame.core.util import array_shift
from static_frame.core.util import array_ufunc_axis_skipna
from static_frame.core.util import arrays_equal
from static_frame.core.util import concat_resolved
from static_frame.core.util import dtype_from_element
from static_frame.core.util import isfalsy_array
from static_frame.core.util import isin
from static_frame.core.util import isna_array
from static_frame.core.util import iterable_to_array_1d
from static_frame.core.util import key_to_str
from static_frame.core.util import pos_loc_slice_to_iloc_slice
from static_frame.core.util import to_datetime64
from static_frame.core.util import ufunc_unique1d_indexer
from static_frame.core.util import validate_dtype_specifier

if tp.TYPE_CHECKING:
    import pandas  # pylint: disable=W0611 #pragma: no cover

    from static_frame import IndexHierarchy  # pylint: disable=W0611 #pragma: no cover
    from static_frame import Series  # pylint: disable=W0611 #pragma: no cover
    from static_frame.core.index_auto import RelabelInput  # pylint: disable=W0611 #pragma: no cover

I = tp.TypeVar('I', bound='Index')


class ILocMeta(type):

    def __getitem__(cls,
            key: GetItemKeyType
            ) -> 'ILoc':
        return cls(key) #type: ignore


class ILoc(metaclass=ILocMeta):
    '''A wrapper for embedding ``iloc`` specifications within a single axis argument of a ``loc`` selection.
    '''

    STATIC = True
    __slots__ = (
            'key',
            )

    def __init__(self, key: GetItemKeyType):
        self.key = key

    def __repr__(self) -> str:
        if isinstance(self.key, tuple):
            return f'<ILoc[{",".join(map(key_to_str, self.key))}]>'
        return f'<ILoc[{key_to_str(self.key)}]>'


def immutable_index_filter(index: IndexBase) -> IndexBase:
    '''Return an immutable index. All index objects handle converting from mutable to immutable via the __init__ constructor; but need to use appropriate class between Index and IndexHierarchy.'''

    if index.STATIC:
        return index
    return index._IMMUTABLE_CONSTRUCTOR(index)


def mutable_immutable_index_filter(
        target_static: bool,
        index: IndexBase,
        ) -> IndexBase:
    if target_static:
        return immutable_index_filter(index)
    # target mutable
    if index.STATIC:
        return index._MUTABLE_CONSTRUCTOR(index)
    return index.__class__(index) # create new instance


#-------------------------------------------------------------------------------

class _ArgsortCache(tp.NamedTuple):
    arr: np.ndarray
    key: np.ndarray

    def __deepcopy__(self, memo: tp.Dict[int, tp.Any]) -> '_ArgsortCache':
        obj = self.__class__(
                array_deepcopy(self.arr, memo=memo),
                array_deepcopy(self.key, memo=memo),
                )

        memo[id(self)] = obj
        return obj


class Index(IndexBase):
    '''A mapping of labels to positions, immutable and of fixed size. Used by default in :obj:`Series` and as index and columns in :obj:`Frame`. Base class of all 1D indices.'''

    __slots__ = (
        '_map',
        '_labels',
        '_positions',
        '_recache',
        '_name',
        '_argsort_cache',
        )

    # _IMMUTABLE_CONSTRUCTOR is None from IndexBase
    # _MUTABLE_CONSTRUCTOR will be set after IndexGO defined

    _DTYPE: tp.Optional[np.dtype] = None # for specialized indices requiring a typed labels

    # for compatability with IndexHierarchy, where this is implemented as a property method
    depth: int = 1
    _NDIM: int = 1

    _map: tp.Optional[FrozenAutoMap]
    _labels: np.ndarray
    _positions: np.ndarray
    _recache: bool
    _name: NameType
    _argsort_cache: tp.Optional[_ArgsortCache]

    #---------------------------------------------------------------------------
    # methods used in __init__ that are customized in derived classes; there, we need to mutate instance state, this these are instance methods
    @staticmethod
    def _extract_labels(
            mapping: tp.Optional[tp.Dict[tp.Hashable, int]],
            labels: tp.Iterable[tp.Hashable],
            dtype: tp.Optional[np.dtype] = None
            ) -> np.ndarray:
        '''Derive labels, a cache of the mapping keys in a sequence type (either an ndarray or a list).

        If the labels passed at instantiation are an ndarray, they are used after immutable filtering. Otherwise, the mapping keys are used to create an ndarray.

        This method is overridden in the derived class.

        Args:
            mapping: Can be None if loc_is_iloc.
            labels: might be an expired Generator, but if it is an immutable ndarray, we can use it without a copy.
        '''
        # pre-fetching labels for faster get_item construction
        if labels.__class__ is np.ndarray:
            if dtype is not None and dtype != labels.dtype: #type: ignore
                raise ErrorInitIndex('invalid label dtype for this Index')
            return immutable_filter(labels)

        # labels may be an expired generator, must use the mapping
        labels_src = labels if hasattr(labels, '__len__') else mapping

        if len(labels_src) == 0: #type: ignore
            if dtype is None:
                labels = EMPTY_ARRAY
            else:
                labels = np.empty(0, dtype=dtype)
                labels.flags.writeable = False #type: ignore
        else: # resolving the dtype is expensive, pass if possible
            labels, _ = iterable_to_array_1d(labels_src, dtype=dtype) #type: ignore

        return labels

    @staticmethod
    def _extract_positions(
            size: int,
            positions: tp.Optional[tp.Sequence[int]]
            ) -> np.ndarray:
        # positions is either None or an ndarray
        if positions.__class__ is np.ndarray:
            return immutable_filter(positions)
        return PositionsAllocator.get(size)

    #---------------------------------------------------------------------------
    # constructors

    @classmethod
    def from_labels(cls: tp.Type[I],
            labels: tp.Iterable[tp.Sequence[tp.Hashable]],
            *,
            name: NameType = None
            ) -> I:
        '''
        Construct an ``Index`` from an iterable of labels, where each label is a hashable. Provided for a compatible interface to ``IndexHierarchy``.
        '''
        return cls(labels, name=name)


    @staticmethod
    def _error_init_index_non_unique(labels: tp.Iterable[tp.Hashable]) -> ErrorInitIndexNonUnique:
        labels_counter = Counter(labels)
        if len(labels_counter) == 0: # generator consumed
            msg = 'Labels have non-unique values. Details from iterators not available.'
        else:
            labels_all = sum(labels_counter.values())
            labels_duplicated = [repr(p[0]) for p in labels_counter.most_common(10) if p[1] > 1]
            msg = f'Labels have {labels_all - len(labels_counter)} non-unique values, including {", ".join(labels_duplicated)}.'

        return ErrorInitIndexNonUnique(msg)

    #---------------------------------------------------------------------------
    @doc_inject(selector='index_init')
    def __init__(self,
            labels: IndexInitializer,
            *,
            loc_is_iloc: bool = False,
            name: NameType = NAME_DEFAULT,
            dtype: DtypeSpecifier = None,
            ) -> None:
        '''Initializer.

        {args}
        '''
        self._recache: bool = False
        self._map: tp.Optional[FrozenAutoMap] = None
        self._argsort_cache: tp.Optional[_ArgsortCache] = None

        positions = None
        is_typed = self._DTYPE is not None # only True for datetime64 indices

        # resolve the targetted labels dtype, by lookin at the class attr _DTYPE and/or the passed dtype argument
        if dtype is None:
            dtype_extract = self._DTYPE # set in some specialized Index sub-classes
        else: # passed dtype is not None
            if is_typed and dtype != self._DTYPE:
                # NOTE: should never get to this branch, as derived Index classes that set _DTYPE remove dtype from __init__
                raise ErrorInitIndex('invalid dtype argument for this Index', dtype, self._DTYPE) #pragma: no cover
            # self._DTYPE is None, passed dtype is not None, use dtype
            dtype_extract = dtype

        #-----------------------------------------------------------------------
        # handle all Index subclasses
        if labels.__class__ is np.ndarray:
            pass
        elif isinstance(labels, IndexBase):
            if labels._recache:
                labels._update_array_cache()
            if name is NAME_DEFAULT:
                name = labels.name # immutable, so no copy necessary

            if labels.depth == 1: # not an IndexHierarchy
                if (labels.STATIC and self.STATIC and dtype is None):
                    if not is_typed or (is_typed and self._DTYPE == labels.dtype):
                        # can take the map if static and if types in the dict are the same as those in the labels (or to become the labels after conversion)
                        self._map = labels._map #type: ignore
                # get a reference to the immutable arrays, even if this is an IndexGO index, we can take the cached arrays, assuming they are up to date; for datetime64 indices, we might need to translate to a different type
                positions = labels._positions #type: ignore
                loc_is_iloc = labels._map is None #type: ignore
                labels = labels._labels # type: ignore
            else: # IndexHierarchy
                # will be a generator of tuples; already updated caches
                labels = labels.__iter__()

        elif isinstance(labels, ContainerOperand):
            # it is a Series or similar
            array = labels.values
            if array.ndim == 1:
                labels = array
            else:
                labels = array2d_to_tuples(array)
        # else: assume an iterable suitable for labels usage, we will identify strings later

        #-----------------------------------------------------------------------
        if is_typed:
            # do not need to check arrays, as will and checked to match dtype_extract in _extract_labels
            if not labels.__class__ is np.ndarray:
                # for now, assume that if _DTYPE is defined, we have a date
                labels = (to_datetime64(v, dtype_extract) for v in labels)
            # coerce to target type
            elif labels.dtype != dtype_extract: #type: ignore
                labels = labels.astype(dtype_extract) #type: ignore
                labels.flags.writeable = False #type: ignore
            labels_for_automap = labels

        self._name = None if name is NAME_DEFAULT else name_filter(name)

        if self._map is None: # if _map not shared from another Index
            if not loc_is_iloc:
                # PERF: calling tolist before initializing AutoMap is shown to be about 2x faster, but can only be done with NumPy dtypes that are equivalent after conversion to Python objects
                if not is_typed:
                    if labels.__class__ is np.ndarray and labels.dtype.kind in DTYPE_OBJECTABLE_KINDS: #type: ignore
                        labels_for_automap = labels.tolist() #type: ignore
                    elif isinstance(labels, str):
                        # NOTE: this is necessary as otherwise a malformed Index will be created, whereby the _map will treat the string as an iterable of chars, while the labels will not and have a single string value. This is consisten as other elements (ints, Booleans) are rejected on instantiation of the AutoMap
                        raise ErrorInitIndex('Cannot create an Index from a single string; provide an iterable of strings.')
                    else:
                        labels_for_automap = labels
                else:
                    labels_for_automap = labels

                try:
                    self._map = FrozenAutoMap(labels_for_automap) if self.STATIC else AutoMap(labels_for_automap)
                except NonUniqueError: # Automap will raise ValueError of non-unique values are encountered
                    raise self._error_init_index_non_unique(labels_for_automap) from None
                # must take length after map as might be iterator
                size = len(self._map)

            else:
                # if loc_is_iloc, labels must be positions and we assume that internal clients that provided loc_is_iloc will not give a generator
                size = len(labels) #type: ignore
                if positions is None:
                    positions = labels
        else: # map shared from another Index
            size = len(self._map)

        # this might be NP array, or a list, depending on if static or grow only; if an array, dtype will be compared with passed dtype_extract
        self._labels = self._extract_labels(self._map, labels, dtype_extract)
        self._positions = self._extract_positions(size, positions)

        if self._DTYPE and self._labels.dtype != self._DTYPE:
            raise ErrorInitIndex('Invalid label dtype for this Index', #pragma: no cover
                    self._labels.dtype, self._DTYPE)

        # NOTE: to implement GH # 374; do this after final self._labels creation as user may pass a dtype argument
        if not is_typed and self._labels.dtype.kind == DTYPE_DATETIME_KIND:
            raise ErrorInitIndex('Cannot create an Index with a datetime64 array; use an Index subclass (e.g. IndexDate) or supply an `index_constructors` argument')

    #---------------------------------------------------------------------------

    def __setstate__(self, state: tp.Tuple[None, tp.Dict[str, tp.Any]]) -> None:
        '''
        Ensure that reanimated NP arrays are set not writeable.
        '''
        for key, value in state[1].items():
            setattr(self, key, value)
        self._labels.flags.writeable = False

    def __deepcopy__(self: I, memo: tp.Dict[int, tp.Any]) -> I:
        assert not self._recache # __deepcopy__ is implemented on derived GO class

        obj = self.__class__.__new__(self.__class__)
        obj._map = deepcopy(self._map, memo)
        obj._labels = array_deepcopy(self._labels, memo)
        obj._positions = PositionsAllocator.get(len(self._labels))
        obj._recache = False
        obj._name = self._name # should be hashable/immutable
        obj._argsort_cache = deepcopy(self._argsort_cache, memo)

        memo[id(self)] = obj
        return obj

    def _memory_label_component_pairs(self,
            ) -> tp.Iterable[tp.Tuple[str, tp.Any]]:
        return (('Name', self._name),
                ('Map', self._map),
                ('Labels', self._labels),
                ('Positions', self._positions),
                )

    def __copy__(self: I) -> I:
        '''
        Return shallow copy of this Index.
        '''
        if self._recache:
            self._update_array_cache()

        return self.__class__(self, name=self._name)

    def copy(self: I) -> I:
        '''
        Return shallow copy of this Index.
        '''
        return self.__copy__()

    #---------------------------------------------------------------------------
    # name interface

    def rename(self: I, name: NameType) -> I:
        '''
        Return a new Frame with an updated name attribute.
        '''
        if self._recache:
            self._update_array_cache()
        # let the constructor handle reuse
        return self.__class__(self, name=name)

    #---------------------------------------------------------------------------
    # interfaces

    @property
    def loc(self) -> InterfaceGetItem[TContainer]:
        return InterfaceGetItem(self._extract_loc) #type: ignore

    @property
    def iloc(self) -> InterfaceGetItem[TContainer]:
        return InterfaceGetItem(self._extract_iloc) #type: ignore

    # # on Index, getitem is an iloc selector; on Series, getitem is a loc selector; for this extraction interface, we do not implement a getitem level function (using iloc would be consistent), as it is better to be explicit between iloc loc

    def _iter_label(self,
            depth_level: tp.Optional[DepthLevelSpecifier] = None
            ) -> tp.Iterator[tp.Hashable]:
        yield from self._labels

    def _iter_label_items(self,
            depth_level: tp.Optional[DepthLevelSpecifier] = None
            ) -> tp.Iterator[tp.Tuple[int, tp.Hashable]]:
        yield from zip(self._positions, self._labels)

    @property
    def iter_label(self) -> IterNodeDepthLevel[tp.Any]:
        return IterNodeDepthLevel(
                container=self,
                function_items=self._iter_label_items,
                function_values=self._iter_label,
                yield_type=IterNodeType.VALUES,
                apply_type=IterNodeApplyType.INDEX_LABELS
                )


    #---------------------------------------------------------------------------
    # common attributes from the numpy array

    @property # type: ignore
    @doc_inject()
    def mloc(self) -> int:
        '''{doc_int}
        '''
        if self._recache:
            self._update_array_cache()
        return mloc(self._labels)

    @property
    def dtype(self) -> np.dtype:
        '''
        Return the dtype of the underlying NumPy array.

        Returns:
            :obj:`numpy.dtype`
        '''
        if self._recache:
            self._update_array_cache()
        return self._labels.dtype

    @property
    def shape(self) -> tp.Tuple[int, ...]:
        '''
        Return a tuple describing the shape of the underlying NumPy array.

        Returns:
            :obj:`tp.Tuple[int]`
        '''
        if self._recache:
            self._update_array_cache()
        return self._labels.shape #type: ignore

    @property
    def ndim(self) -> int:
        '''
        Return the number of dimensions.

        Returns:
            :obj:`int`
        '''
        if self._recache:
            self._update_array_cache()
        return self._labels.ndim #type: ignore

    @property
    def size(self) -> int:
        '''
        Return the size of the underlying NumPy array.

        Returns:
            :obj:`int`
        '''
        if self._recache:
            self._update_array_cache()
        return self._labels.size #type: ignore

    @property
    def nbytes(self) -> int:
        '''
        Return the total bytes of the underlying NumPy array.

        Returns:
            :obj:`int`
        '''
        if self._recache:
            self._update_array_cache()
        return self._labels.nbytes #type: ignore

    #---------------------------------------------------------------------------
    def _drop_iloc(self, key: GetItemKeyType) -> 'Index':
        '''Create a new index after removing the values specified by the iloc key.
        '''
        if self._recache:
            self._update_array_cache()

        if key is None:
            if self.STATIC: # immutable, no selection, can return self
                return self
            labels = self._labels # already immutable
        elif key.__class__ is np.ndarray and key.dtype == bool: #type: ignore
            # can use labels, as we already recached
            # use Boolean area to select indices from positions, as np.delete does not work with arrays
            labels = np.delete(self._labels, self._positions[key], axis=0)
            labels.flags.writeable = False
        else:
            labels = np.delete(self._labels, key, axis=0)
            labels.flags.writeable = False

        # from labels will work with both Index and IndexHierarchy
        return self.__class__.from_labels(labels, name=self._name)

    def _drop_loc(self, key: GetItemKeyType) -> 'IndexBase':
        '''Create a new index after removing the values specified by the loc key.
        '''
        return self._drop_iloc(self._loc_to_iloc(key))


    @property
    def drop(self) -> InterfaceSelectDuo[TContainer]:
        return InterfaceSelectDuo( #type: ignore
            func_iloc=self._drop_iloc,
            func_loc=self._drop_loc,
            )


    @doc_inject(select='astype')
    def astype(self, dtype: DtypeSpecifier) -> 'Index':
        '''
        Return an Index with type determined by `dtype` argument. If a `datetime64` dtype is provided, the appropriate ``Index`` subclass will be returned. Note that for Index, this is a simple function, whereas for ``IndexHierarchy``, this is an interface exposing both a callable and a getitem interface.

        Args:
            {dtype}
        '''
        from static_frame.core.index_datetime import dtype_to_index_cls

        dtype = validate_dtype_specifier(dtype)

        array = self.values.astype(dtype)
        array.flags.writeable = False
        cls = dtype_to_index_cls(self.STATIC, array.dtype)
        return cls(
                array,
                name=self._name
                )


    #---------------------------------------------------------------------------

    @property
    def via_values(self) -> InterfaceValues['Index']:
        '''
        Interface for applying functions to values (as arrays) in this container.
        '''
        if self._recache:
            self._update_array_cache()

        return InterfaceValues(self)

    @property
    def via_str(self) -> InterfaceString[np.ndarray]:
        '''
        Interface for applying string methods to elements in this container.
        '''
        if self._recache:
            self._update_array_cache()

        def blocks_to_container(blocks: tp.Iterator[np.ndarray]) -> np.ndarray:
            return next(blocks)

        return InterfaceString(
                blocks=(self._labels,),
                blocks_to_container=blocks_to_container,
                ndim=self._NDIM,
                labels=range(1)
                )

    @property
    def via_dt(self) -> InterfaceDatetime[np.ndarray]:
        '''
        Interface for applying datetime properties and methods to elements in this container.
        '''
        if self._recache:
            self._update_array_cache()

        def blocks_to_container(blocks: tp.Iterator[np.ndarray]) -> np.ndarray:
            return next(blocks)

        return InterfaceDatetime(
                blocks=(self.values,),
                blocks_to_container=blocks_to_container,
                )

    def via_re(self,
            pattern: str,
            flags: int = 0,
            ) -> InterfaceRe[np.ndarray]:
        '''
        Interface for applying regular expressions to elements in this container.
        '''
        if self._recache:
            self._update_array_cache()

        def blocks_to_container(blocks: tp.Iterator[np.ndarray]) -> np.ndarray:
            return next(blocks)

        return InterfaceRe(
                blocks=(self._labels,),
                blocks_to_container=blocks_to_container,
                pattern=pattern,
                flags=flags,
                )

    #---------------------------------------------------------------------------

    def _update_array_cache(self) -> None:
        '''Derived classes can use this to set stored arrays, self._labels and self._positions.
        '''

    #---------------------------------------------------------------------------

    def __len__(self) -> int:
        if self._recache:
            self._update_array_cache()
        return len(self._labels)

    @doc_inject()
    def display(self,
            config: tp.Optional[DisplayConfig] = None,
            *,
            style_config: tp.Optional[StyleConfig] = None,
            ) -> Display:
        '''{doc}

        Args:
            {config}
        '''
        config = config or DisplayActive.get()

        if self._recache:
            self._update_array_cache()

        header: tp.Optional[DisplayHeader]

        if config.type_show:
            header = DisplayHeader(self.__class__, self._name)
            header_depth = 1
        else:
            header = None
            header_depth = 0

        return Display.from_values(self.values,
                header=header,
                config=config,
                outermost=True,
                index_depth=0,
                header_depth=header_depth,
                style_config=style_config,
                )

    #---------------------------------------------------------------------------
    # core internal representation

    @property #type: ignore
    @doc_inject(selector='values_1d', class_name='Index')
    def values(self) -> np.ndarray:
        '''
        {}
        '''
        if self._recache:
            self._update_array_cache()
        return self._labels

    @property
    def positions(self) -> np.ndarray:
        '''Return the immutable positions array.
        '''
        # This is needed by some clients, such as Series and Frame, to support Boolean usage in drop.
        if self._recache:
            self._update_array_cache()
        return self._positions

    def _get_argsort_cache(self: I) -> _ArgsortCache:
        '''
        Return a cached NT containing self.values sorted, along with the argsort key

        This utilizes a lazy instance cache attribute, since sorting is expensive,
        and this operation is typically called either never, or often.
        '''
        if self._recache:
            self._update_array_cache()

        if self._argsort_cache is None:
            self._argsort_cache = _ArgsortCache(*ufunc_unique1d_indexer(self.values))

        return self._argsort_cache

    def _index_iloc_map(self: I, other: I) -> np.ndarray:
        '''
        Return an array of index locations to map from this array to another

        Equivalent to: self.iter_label().apply(other._loc_to_iloc)
        '''
        if self.__len__() == 0:
            return EMPTY_ARRAY

        # Equivalent to: ufunc_unique1d_indexer(self.values)
        ar1, ar1_indexer = self._get_argsort_cache()
        ar2 = other.values

        aux = concat_resolved((ar1, ar2))
        aux_sort_indices = argsort_array(aux)
        aux = aux[aux_sort_indices]

        mask = aux[1:] == aux[:-1]

        indexer = aux_sort_indices[1:][mask] - ar1.size

        # We want to return these indices to match ar1 before it was sorted
        try:
            indexer = indexer[ar1_indexer]
        except IndexError as e:
            # Display the first missing element
            raise KeyError(self.difference(other)[0]) from e

        indexer.flags.writeable = False
        return indexer

    @staticmethod
    def _depth_level_validate(depth_level: DepthLevelSpecifier) -> None:
        '''
        Handle all variety of depth_level specifications for a 1D index: only 0, -1, and lists of the same are valid.
        '''
        if not isinstance(depth_level, INT_TYPES):
            depth_level = tuple(depth_level)
            if len(depth_level) != 1:
                raise RuntimeError('invalid depth_level', depth_level)
            depth_level = depth_level[0]

        if depth_level > 0 or depth_level < -1:
            raise RuntimeError('invalid depth_level', depth_level)

    def values_at_depth(self,
            depth_level: DepthLevelSpecifier = 0
            ) -> np.ndarray:
        '''
        Return an NP array for the `depth_level` specified.
        '''
        self._depth_level_validate(depth_level)
        return self.values

    @doc_inject()
    def label_widths_at_depth(self,
            depth_level: DepthLevelSpecifier = 0
            ) -> tp.Iterator[tp.Tuple[tp.Hashable, int]]:
        '''{}'''
        self._depth_level_validate(depth_level)
        yield from zip_longest(self.values, (), fillvalue=1)

    @property
    def index_types(self) -> 'Series':
        '''
        Return a Series of Index classes for each index depth.

        Returns:
            :obj:`Series`
        '''
        from static_frame.core.series import Series
        return Series((self.__class__,), index=(self._name,), dtype=DTYPE_OBJECT)


    #---------------------------------------------------------------------------

    def relabel(self, mapper: 'RelabelInput') -> 'Index':
        '''
        Return a new Index with labels replaced by the callable or mapping; order will be retained. If a mapping is used, the mapping need not map all origin keys.
        '''
        if self._recache:
            self._update_array_cache()

        if not callable(mapper):
            # if a mapper, it must support both __getitem__ and __contains__
            getitem = getattr(mapper, '__getitem__')
            return self.__class__(
                    (getitem(x) if x in mapper else x for x in self._labels),
                    name=self._name
                    )

        return self.__class__(
                (mapper(x) for x in self._labels),
                name=self._name
                )

    #---------------------------------------------------------------------------
    # extraction and selection

    def _loc_to_iloc(self,
            key: GetItemKeyType,
            key_transform: KeyTransformType = None,
            partial_selection: bool = False,
            ) -> GetItemKeyType:
        '''
        Args:
            key_transform: A function that transforms keys to specialized type; used by IndexDate indices.
        Returns:
            Return GetItemKey type that is based on integers, compatible with TypeBlocks
        '''
        if key.__class__ is ILoc:
            return key.key #type: ignore

        key = key_from_container_key(self, key)

        if self._map is None: # loc_is_iloc
            if key.__class__ is np.ndarray:
                if key.dtype == DTYPE_BOOL: #type: ignore
                    return key
                if key.dtype != DTYPE_INT_DEFAULT: #type: ignore
                    # if key is an np.array, it must be an int or bool type
                    # could use tolist(), but we expect all keys to be integers
                    return key.astype(DTYPE_INT_DEFAULT) #type: ignore
            elif key.__class__ is slice:
                # might raise LocInvalid
                key = pos_loc_slice_to_iloc_slice(key, self.__len__())
            return key

        if key_transform:
            key = key_transform(key)

        # PERF: isolate for usage of _positions
        if self._recache:
            self._update_array_cache()

        return LocMap.loc_to_iloc(
                label_to_pos=self._map,
                labels=self._labels,
                positions=self._positions, # always an np.ndarray
                key=key,
                partial_selection=partial_selection,
                )

    def loc_to_iloc(self,
            key: GetItemKeyType,
            ) -> GetItemKeyType:
        '''Given a label (loc) style key (either a label, a list of labels, a slice, or a Boolean selection), return the index position (iloc) style key. Keys that are not found will raise a KeyError or a sf.LocInvalid error.

        Args:
            key: a label key.
        '''
        if self._map is None: # loc is iloc
            # NOTE: the specialization here is to use the key on the positions array and return iloc values, rather than just propagating the selection array. This also handles and re-raises better exceptions.

            if not key.__class__ is slice:
                if self._recache:
                    self._update_array_cache()

                key = key_from_container_key(self, key)
                is_array = key.__class__ is np.ndarray
                try:
                    # NOTE: this insures that the returned type will be DTYPE_INT_DEFAULT
                    result = self._positions[key]
                except IndexError as e:
                    # NP gives us: IndexError: only integers, slices (`:`), ellipsis (`...`), numpy.newaxis (`None`) and integer or boolean arrays are valid indices
                    if is_array and key.dtype == DTYPE_BOOL: #type: ignore
                        raise # loc selection on Boolean array selection returns IndexError
                    raise KeyError(key) from e

                return result # return position as array

            # might raise LocInvalid
            return pos_loc_slice_to_iloc_slice(key, self.__len__())

        return self._loc_to_iloc(key)

    def _extract_iloc(self,
            key: GetItemKeyType,
            ) -> tp.Union['Index', tp.Hashable]:
        '''Extract a new index given an iloc key.
        '''
        if self._recache:
            self._update_array_cache()

        if key is None:
            labels = self._labels
            loc_is_iloc = self._map is None
        elif key.__class__ is slice:
            if key == NULL_SLICE:
                labels = self._labels
                loc_is_iloc = self._map is None
            else:
                # if labels is an np array, this will be a view; if a list, a copy
                labels = self._labels[key]
                labels.flags.writeable = False
                loc_is_iloc = False
        elif isinstance(key, KEY_ITERABLE_TYPES):
            # can select directly from _labels[key] if if key is a list, array, or Boolean array
            labels = self._labels[key]
            labels.flags.writeable = False
            loc_is_iloc = False
        else: # select a single label value
            return self._labels[key] #type: ignore

        return self.__class__(labels=labels,
                loc_is_iloc=loc_is_iloc,
                name=self._name,
                )

    def _extract_iloc_by_int(self,
            key: int,
            ) -> tp.Hashable:
        '''Extract an element given an iloc integer key.
        '''
        if self._recache:
            self._update_array_cache()
        return self._labels[key] #type: ignore

    def _extract_loc(self: I,
            key: GetItemKeyType
            ) -> tp.Union['Index', tp.Hashable]:
        return self._extract_iloc(self._loc_to_iloc(key))

    def __getitem__(self: I,
            key: GetItemKeyType
            ) -> tp.Union['Index', tp.Hashable]:
        '''Extract a new index given an iloc key.
        '''
        return self._extract_iloc(key)

    #---------------------------------------------------------------------------
    # operators

    def _ufunc_unary_operator(self,
            operator: UFunc
            ) -> np.ndarray:
        '''Always return an NP array.
        '''
        if self._recache:
            self._update_array_cache()

        array = operator(self._labels)
        array.flags.writeable = False
        return array

    def _ufunc_binary_operator(self, *,
            operator: UFunc,
            other: tp.Any,
            fill_value: object = np.nan,
            ) -> np.ndarray:
        '''
        Binary operators applied to an index always return an NP array. This deviates from Pandas, where some operations (multiplying an int index by an int) result in a new Index, while other operations result in a np.array (using == on two Index).
        '''
        from static_frame.core.frame import Frame
        from static_frame.core.series import Series

        if self._recache:
            self._update_array_cache()

        if isinstance(other, (Series, Frame)):
            raise ValueError('cannot use labelled container as an operand.')

        values = self._labels
        other_is_array = False

        if issubclass(other.__class__, Index):
            other = other.values # operate on labels to labels
            other_is_array = True
        elif other.__class__ is np.ndarray:
            other_is_array = True

        if operator.__name__ == 'matmul':
            return matmul(values, other)
        elif operator.__name__ == 'rmatmul':
            return matmul(other, values)

        return apply_binary_operator(
                values=values,
                other=other,
                other_is_array=other_is_array,
                operator=operator,
                )

    def _ufunc_axis_skipna(self, *,
            axis: int,
            skipna: bool,
            ufunc: UFunc,
            ufunc_skipna: UFunc,
            composable: bool,
            dtypes: tp.Tuple[np.dtype, ...],
            size_one_unity: bool
            ) -> np.ndarray:
        '''

        Args:
            dtype: Not used in 1D application, but collected here to provide a uniform signature.
        '''
        if self._recache:
            self._update_array_cache()

        # do not need to pass on composabel here
        return array_ufunc_axis_skipna(
                array=self._labels,
                skipna=skipna,
                axis=0,
                ufunc=ufunc,
                ufunc_skipna=ufunc_skipna
                )

    def _ufunc_shape_skipna(self, *,
            axis: int,
            skipna: bool,
            ufunc: UFunc,
            ufunc_skipna: UFunc,
            composable: bool,
            dtypes: tp.Tuple[np.dtype, ...],
            size_one_unity: bool
            ) -> np.ndarray:
        '''
        As Index and IndexHierarchy return np.ndarray from such operations, _ufunc_shape_skipna and _ufunc_axis_skipna can be defined the same.

        Returns:
            immutable NumPy array.
        '''
        # NOTE: for 1D Index, can use axis for shape ufunc
        return self._ufunc_axis_skipna(
                axis=axis,
                skipna=skipna,
                ufunc=ufunc,
                ufunc_skipna=ufunc_skipna,
                composable=composable, # shape on axis 1 is never composable
                dtypes=dtypes,
                size_one_unity=size_one_unity
                )

    #---------------------------------------------------------------------------
    # dictionary-like interface

    # NOTE: we intentionally exclude keys(), items(), and get() from Index classes, as they return inconsistent result when thought of as a dictionary


    def __iter__(self) -> tp.Iterator[tp.Hashable]:
        '''Iterate over labels.
        '''
        if self._recache:
            self._update_array_cache()
        yield from self._labels.__iter__()

    def __reversed__(self) -> tp.Iterator[tp.Hashable]:
        '''
        Returns a reverse iterator on the index labels.
        '''
        if self._recache:
            self._update_array_cache()
        return reversed(self._labels)

    def __contains__(self, value: tp.Any) -> bool:
        '''Return True if value in the labels.
        '''
        if self._map is None: # loc_is_iloc
            if isinstance(value, INT_TYPES):
                return value >= 0 and value < len(self) #type: ignore
            return False #type: ignore [unreachable]
        return self._map.__contains__(value) #type: ignore


    #---------------------------------------------------------------------------
    # utility functions

    def unique(self,
            depth_level: DepthLevelSpecifier = 0,
            order_by_occurrence: bool = False,
            ) -> np.ndarray:
        '''
        Return a NumPy array of unique values.

        Args:
            depth_level: defaults to 0 for for a 1D Index.
            order_by_occurrence: for 1D indices, this argument is a no-op. Provided for compatibility with IndexHierarchy.

        Returns:
            :obj:`numpy.ndarray`
        '''
        self._depth_level_validate(depth_level)
        return self.values

    @doc_inject()
    def equals(self,
            other: tp.Any,
            *,
            compare_name: bool = False,
            compare_dtype: bool = False,
            compare_class: bool = False,
            skipna: bool = True,
            ) -> bool:
        '''
        {doc}

        Args:
            {compare_name}
            {compare_dtype}
            {compare_class}
            {skipna}
        '''

        if id(other) == id(self):
            return True

        if compare_class and self.__class__ != other.__class__:
            return False
        elif not isinstance(other, Index):
            return False

        # defer updating cache
        if self._recache:
            self._update_array_cache()

        # same type from here
        if len(self) != len(other):
            return False
        if compare_name and self.name != other.name:
            return False
        if self._map is None and other._map is None:
            return True # have same length must be same integer range and dtype
        if compare_dtype and self.dtype != other.dtype:
            return False
        return arrays_equal(self.values, other.values, skipna=skipna)


    @doc_inject(selector='sort')
    def sort(self,
            ascending: bool = True,
            kind: str = DEFAULT_SORT_KIND,
            key: tp.Optional[tp.Callable[['Index'], tp.Union[np.ndarray, 'Index']]] = None,
            ) -> 'Index':
        '''Return a new Index with the labels sorted.

        Args:
            {ascending}
            {kind}
            {key}
        '''
        order = sort_index_for_order(self, kind=kind, ascending=ascending, key=key) #type: ignore [arg-type]

        return self._extract_iloc(order) #type: ignore [return-value]

    def isin(self, other: tp.Iterable[tp.Any]) -> np.ndarray:
        '''
        Return a Boolean array showing True where a label is found in other. If other is a multidimensional array, it is flattened.
        '''
        return isin(self.values, other, array_is_unique=True)

    def roll(self, shift: int) -> 'Index':
        '''Return an Index with values rotated forward and wrapped around (with a postive shift) or backward and wrapped around (with a negative shift).
        '''
        values = self.values # force usage of property for cache update
        if shift % len(values):
            values = array_shift(
                    array=values,
                    shift=shift,
                    axis=0,
                    wrap=True)
            values.flags.writeable = False
        return self.__class__(values, name=self._name)

    #---------------------------------------------------------------------------
    # na handling
    # falsy handling

    def _drop_missing(self,
            func: tp.Callable[[np.ndarray], np.ndarray],
            dtype_kind_targets: tp.Optional[tp.FrozenSet[str]],
            ) -> 'Index':
        '''
        Args:
            func: UFunc that returns True for missing values
        '''
        labels = self.values
        if dtype_kind_targets is not None and labels.dtype.kind not in dtype_kind_targets:
            return self if self.STATIC else self.copy()

        # get positions that we want to keep
        isna = func(labels)
        length = len(labels)
        count = isna.sum()

        if count == length: # all are NaN
            return self.__class__((), name=self.name)
        if count == 0: # None are nan
            return self if self.STATIC else self.copy()

        sel = np.logical_not(isna)
        values = labels[sel]
        values.flags.writeable = False

        return self.__class__(values,
                name=self._name,
                )

    def dropna(self) -> 'Index':
        '''
        Return a new :obj:`Index` after removing values of NaN or None.
        '''
        return self._drop_missing(isna_array, DTYPE_NA_KINDS)

    def dropfalsy(self) -> 'Index':
        '''
        Return a new :obj:`Index` after removing values of NaN or None.
        '''
        return self._drop_missing(isfalsy_array, None)

    #---------------------------------------------------------------------------

    def _fill_missing(self,
            func: tp.Callable[[np.ndarray], np.ndarray],
            value: tp.Any,
            ) -> 'Index':
        values = self.values # force usage of property for cache update
        sel = func(values)
        if not np.any(sel):
            return self if self.STATIC else self.copy()

        value_dtype = dtype_from_element(value)
        assignable_dtype = resolve_dtype(value_dtype, values.dtype)

        if values.dtype == assignable_dtype:
            assigned = values.copy()
        else:
            assigned = values.astype(assignable_dtype)

        assigned[sel] = value
        assigned.flags.writeable = False
        return self.__class__(assigned, name=self._name)

    @doc_inject(selector='fillna')
    def fillna(self, value: tp.Any) -> 'Index':
        '''Return an :obj:`Index` with replacing null (NaN or None) with the supplied value.

        Args:
            {value}
        '''
        return self._fill_missing(isna_array, value)

    @doc_inject(selector='fillna')
    def fillfalsy(self, value: tp.Any) -> 'Index':
        '''Return an :obj:`Index` with replacing falsy values with the supplied value.

        Args:
            {value}
        '''
        return self._fill_missing(isfalsy_array, value)

    #---------------------------------------------------------------------------
    def _sample_and_key(self,
            count: int = 1,
            *,
            seed: tp.Optional[int] = None,
            ) -> tp.Tuple['Index', np.ndarray]:
        # NOTE: base class defines pubic method
        # force usage of property for cache update
        # sort positions to avoid uncomparable objects
        key = array_sample(self.positions, count=count, seed=seed, sort=True)

        values = self.values[key]
        values.flags.writeable = False
        return self.__class__(values, name=self._name), key

    def level_add(self,
            level: tp.Hashable,
            *,
            index_constructor: IndexConstructor = None,
            ) -> 'IndexHierarchy':
        '''Return an IndexHierarchy with an added root level.

        Args:
            level: A hashable to used as the new root.
            *,
            index_constructor:
        '''
        from static_frame import Index
        from static_frame import IndexGO
        from static_frame import IndexHierarchy
        from static_frame import IndexHierarchyGO

        cls = IndexHierarchy if self.STATIC else IndexHierarchyGO
        cls_depth = Index if self.STATIC else IndexGO

        if index_constructor is None:
            # cannot assume new depth is the same index subclass
            index_constructor = cls_depth

        indices = [index_constructor((level,)), immutable_index_filter(self)]

        indexers = np.array(
                [
                    np.zeros(self.__len__(), dtype=DTYPE_INT_DEFAULT),
                    PositionsAllocator.get(self.__len__())
                ]
        )
        indexers.flags.writeable = False

        return cls(
                indices=indices,
                indexers=indexers,
                name=self._name,
                )

    #---------------------------------------------------------------------------
    # export

    def to_series(self) -> 'Series':
        '''Return a Series with values from this Index's labels.
        '''
        # NOTE: while we might re-use the index on the index returned from this Series, such an approach will not work with IndexHierarchy.to_frame, as we do not know if the index should be on the index or columns; thus, returning an unindexed Series is appropriate
        from static_frame import Series
        return Series(self.values, name=self._name)

    def to_pandas(self) -> 'pandas.Index':
        '''Return a Pandas Index.
        '''
        import pandas

        # must copy to remove immutability, decouple reference
        if self._map is None:
            return pandas.RangeIndex(self.__len__(), name=self._name)
        return pandas.Index(self.values.copy(),
                name=self._name)


    def _to_signature_bytes(self,
            include_name: bool = True,
            include_class: bool = True,
            encoding: str = 'utf-8',
            ) -> bytes:

        return b''.join(chain(
                iter_component_signature_bytes(self,
                        include_name=include_name,
                        include_class=include_class,
                        encoding=encoding),
                (self.values.tobytes(),),
                ))

#-------------------------------------------------------------------------------

class _IndexGOMixin:

    STATIC = False
    # NOTE: must define __slots__ in derived class or get TypeError: multiple bases have instance lay-out conflict
    __slots__ = ()

    _map: tp.Optional[AutoMap]
    _labels: np.ndarray
    _positions: np.ndarray
    _labels_mutable: tp.List[tp.Hashable]
    _labels_mutable_dtype: np.dtype
    _positions_mutable_count: int
    _argsort_cache: tp.Optional[_ArgsortCache]

    #---------------------------------------------------------------------------
    def __deepcopy__(self: I, memo: tp.Dict[int, tp.Any]) -> I: #type: ignore
        if self._recache:
            self._update_array_cache()

        obj = self.__class__.__new__(self.__class__)
        obj._map = deepcopy(self._map, memo)
        obj._labels = array_deepcopy(self._labels, memo)
        obj._positions = PositionsAllocator.get(len(self._labels))
        obj._recache = False # pylint: disable=E0237
        obj._name = self._name # pylint: disable=E0237
        obj._labels_mutable = deepcopy(self._labels_mutable, memo) #type: ignore
        obj._labels_mutable_dtype = deepcopy(self._labels_mutable_dtype, memo) #type: ignore
        obj._positions_mutable_count = self._positions_mutable_count #type: ignore
        obj._argsort_cache = deepcopy(self._argsort_cache, memo)

        memo[id(self)] = obj
        return obj

    #---------------------------------------------------------------------------
    def _extract_labels(self,
            mapping: tp.Optional[tp.Dict[tp.Hashable, int]],
            labels: np.ndarray,
            dtype: tp.Optional[np.dtype] = None
            ) -> np.ndarray:
        '''Called in Index.__init__(). This creates and populates mutable storage as a side effect of array derivation; this storage will be grown as needed.
        '''
        labels = Index._extract_labels(mapping, labels, dtype)
        self._labels_mutable = labels.tolist()
        if len(labels):
            self._labels_mutable_dtype = labels.dtype
        else: # avoid setting to float default when labels is empty
            self._labels_mutable_dtype = None
        return labels

    def _extract_positions(self,
            size: int,
            positions: tp.Optional[tp.Sequence[int]]
            ) -> np.ndarray:
        '''Called in Index.__init__(). This creates and populates mutable storage. This creates and populates mutable storage as a side effect of array derivation.
        '''
        positions = Index._extract_positions(size, positions)
        self._positions_mutable_count = size
        return positions

    def _update_array_cache(self) -> None:

        if self._labels_mutable_dtype is not None and len(self._labels):
            # only update if _labels_mutable_dtype has been set and _labels exist
            self._labels_mutable_dtype = resolve_dtype(
                    self._labels.dtype,
                    self._labels_mutable_dtype)

        # NOTE: necessary to support creation from iterable of tuples
        self._labels, _ = iterable_to_array_1d(
                self._labels_mutable,
                dtype=self._labels_mutable_dtype)
        self._positions = PositionsAllocator.get(self._positions_mutable_count)
        self._recache = False # pylint: disable=E0237

        # clear cache
        self._argsort_cache = None

    #---------------------------------------------------------------------------
    # grow only mutation

    def append(self, value: tp.Hashable) -> None:
        '''append a value
        '''
        if self.__contains__(value): #type: ignore
            raise KeyError(f'duplicate key append attempted: {value}')

        # we might need to initialize map if not an increment that keeps loc_is_iloc relationship
        initialize_map = False
        if self._map is None: # loc_is_iloc
            if not (isinstance(value, INT_TYPES)
                    and value == self._positions_mutable_count):
                initialize_map = True
        else:
            self._map.add(value)

        if self._labels_mutable_dtype is not None:
            self._labels_mutable_dtype = resolve_dtype(
                    dtype_from_element(value),
                    self._labels_mutable_dtype)
        else:
            self._labels_mutable_dtype = dtype_from_element(value)

        self._labels_mutable.append(value)

        if initialize_map:
            self._map = AutoMap(self._labels_mutable)

        self._positions_mutable_count += 1
        self._recache = True # pylint: disable=E0237

    def extend(self, values: KeyIterableTypes) -> None:
        '''Append multiple values
        Args:
            values: can be a generator.
        '''
        for value in values:
            self.append(value)


INDEX_GO_LEAF_SLOTS = (
        '_labels_mutable',
        '_labels_mutable_dtype',
        '_positions_mutable_count',
        )

class IndexGO(_IndexGOMixin, Index):
    '''A mapping of labels to positions, immutable with grow-only size. Used as columns in :obj:`FrameGO`.
    '''

    _IMMUTABLE_CONSTRUCTOR = Index
    __slots__ = INDEX_GO_LEAF_SLOTS


# update class attr on Index after class initialziation
Index._MUTABLE_CONSTRUCTOR = IndexGO



#-------------------------------------------------------------------------------

def _index_initializer_needs_init(
        value: tp.Optional[IndexInitializer]
        ) -> bool:
    '''Determine if value is a non-empty index initializer. This could almost just be a truthy test, but ndarrays need to be handled in isolation. Generators should return True.
    '''
    if value is None:
        return False
    if isinstance(value, IndexBase):
        return False
    if value.__class__ is np.ndarray:
        return bool(len(value)) #type: ignore
    return bool(value)




import typing as tp

import numpy as np

from static_frame.core.node_selector import Interface
from static_frame.core.node_selector import InterfaceBatch
from static_frame.core.node_selector import InterfaceGetItem
from static_frame.core.util import KEY_MULTIPLE_TYPES
from static_frame.core.util import NULL_SLICE
from static_frame.core.util import OPERATORS
from static_frame.core.util import AnyCallable
from static_frame.core.util import GetItemKeyType
from static_frame.core.util import GetItemKeyTypeCompound

if tp.TYPE_CHECKING:
    from static_frame.core.batch import Batch  # pylint: disable = W0611 #pragma: no cover
    from static_frame.core.frame import Frame  # pylint: disable = W0611 #pragma: no cover
    from static_frame.core.frame import FrameGO  # pylint: disable = W0611 #pragma: no cover
    from static_frame.core.index import Index  # pylint: disable = W0611 #pragma: no cover
    from static_frame.core.index_base import IndexBase  # pylint: disable = W0611 #pragma: no cover
    from static_frame.core.index_hierarchy import IndexHierarchy  # pylint: disable = W0611 #pragma: no cover
    from static_frame.core.node_transpose import InterfaceBatchTranspose  # pylint: disable = W0611 #pragma: no cover
    from static_frame.core.node_transpose import InterfaceTranspose  # pylint: disable = W0611 #pragma: no cover
    from static_frame.core.series import Series  # pylint: disable = W0611 #pragma: no cover
    from static_frame.core.type_blocks import TypeBlocks  # pylint: disable = W0611 #pragma: no cover

TContainer = tp.TypeVar('TContainer',
        'Frame',
        'Series',
        )
INTERFACE_FILL_VALUE = (
        'loc',
        '__getitem__',
        'via_T',
        '__add__',
        '__sub__',
        '__mul__',
        # '__matmul__',
        '__truediv__',
        '__floordiv__',
        '__mod__',
        '__pow__',
        '__lshift__',
        '__rshift__',
        '__and__',
        '__xor__',
        '__or__',
        '__lt__',
        '__le__',
        '__eq__',
        '__ne__',
        '__gt__',
        '__ge__',
        '__radd__',
        '__rsub__',
        '__rmul__',
        # '__rmatmul__',
        '__rtruediv__',
        '__rfloordiv__',
        )


class InterfaceFillValue(Interface[TContainer]):

    __slots__ = (
            '_container',
            '_fill_value',
            '_axis',
            )

    INTERFACE = INTERFACE_FILL_VALUE

    def __init__(self,
            container: TContainer,
            *,
            fill_value: tp.Any = np.nan,
            axis: int = 0,
            ) -> None:
        self._container: TContainer = container
        self._fill_value = fill_value
        self._axis = axis

    #---------------------------------------------------------------------------
    @property
    def via_T(self) -> 'InterfaceTranspose[Frame]':
        '''
        Interface for using binary operators with one-dimensional sequences, where the opperand is applied column-wise.
        '''
        from static_frame.core.frame import Frame
        from static_frame.core.node_transpose import InterfaceTranspose
        if not isinstance(self._container, Frame):
            raise NotImplementedError('via_T functionality only available on Frame')
        return InterfaceTranspose(
                container=self._container,
                fill_value=self._fill_value,
                )

    #---------------------------------------------------------------------------
    @staticmethod
    def _extract_key_attrs(
            key: GetItemKeyType,
            index: 'IndexBase',
            ) -> tp.Tuple[GetItemKeyType, bool, bool]:
        key_is_multiple = isinstance(key, KEY_MULTIPLE_TYPES)
        if key.__class__ is slice:
            key_is_null_slice = key == NULL_SLICE
            key = index._extract_loc(key) #type: ignore
        else:
            key_is_null_slice = False
        return key, key_is_multiple, key_is_null_slice

    def _extract_loc1d(self,
            key: GetItemKeyType = NULL_SLICE,
            ) -> 'Series':
        '''This is only called if container is 1D
        '''
        from static_frame.core.container_util import get_col_fill_value_factory

        key, is_multiple, is_null_slice = self._extract_key_attrs(
                key,
                self._container._index,
                )
        fill_value = self._fill_value
        container = self._container

        if is_multiple:
            return container.reindex(key if not is_null_slice else None, #type: ignore
                    fill_value=fill_value,
                    )

        fv = get_col_fill_value_factory(fill_value, None)(0, container.dtype) #type: ignore
        return container.get(key, fv) #type: ignore

    def _extract_loc2d(self,
            row_key: GetItemKeyType = NULL_SLICE,
            column_key: GetItemKeyType = NULL_SLICE,
            ) -> tp.Union['Frame', 'Series']:
        '''
        NOTE: keys are loc keys; None is interpreted as selector, not a NULL_SLICE
        '''
        from static_frame.core.container_util import get_col_fill_value_factory
        from static_frame.core.series import Series

        fill_value = self._fill_value
        container = self._container # always a Frame

        row_key, row_is_multiple, row_is_null_slice = self._extract_key_attrs(
                row_key,
                container._index,
                )
        column_key, column_is_multiple, column_is_null_slice = self._extract_key_attrs(
                column_key,
                container._columns, #type: ignore
                )

        if row_is_multiple and column_is_multiple:
            # cannot reindex if loc keys are elements
            return container.reindex( # type: ignore
                    index=row_key if not row_is_null_slice else None,
                    columns=column_key if not column_is_null_slice else None,
                    fill_value=fill_value,
                    )
        elif not row_is_multiple and not column_is_multiple: # selecting an element
            try:
                return container.loc[row_key, column_key]
            except KeyError:
                fv = get_col_fill_value_factory(fill_value, None)(0, None)
                return fv #type: ignore
        elif not row_is_multiple:
            # row is an element, return Series indexed by columns
            if row_key in container._index: #type: ignore
                s = container.loc[row_key]
                return s.reindex(column_key, fill_value=fill_value) #type: ignore
            fv = get_col_fill_value_factory(fill_value, None)(0, None)
            return Series.from_element(fv,
                    index=column_key,
                    name=row_key,
                    )
        # columns is an element, return Series indexed by index
        if column_key in container._columns: #type: ignore
            s = container[column_key]
            return s.reindex(row_key, fill_value=fill_value) #type: ignore

        fv = get_col_fill_value_factory(fill_value, None)(0, None)
        return Series.from_element(fv,
                index=row_key,
                name=column_key,
                )

    def _extract_loc2d_compound(self, key: GetItemKeyTypeCompound) -> tp.Union['Frame', 'Series']:
        if isinstance(key, tuple):
            row_key, column_key = key
        else:
            row_key = key
            column_key = NULL_SLICE
        return self._extract_loc2d(row_key, column_key)

    #---------------------------------------------------------------------------
    @property
    def loc(self) -> InterfaceGetItem['Frame']:
        '''Label-based selection where labels not specified will define a new container containing those labels filled with the fill value.
        '''
        if self._container._NDIM == 1:
            return InterfaceGetItem(self._extract_loc1d)
        return InterfaceGetItem(self._extract_loc2d_compound)

    def __getitem__(self,  key: GetItemKeyType) -> tp.Union['Frame', 'Series']:
        '''Label-based selection where labels not specified will define a new container containing those labels filled with the fill value.
        '''
        if self._container._NDIM == 1:
            return self._extract_loc1d(key)
        return self._extract_loc2d(NULL_SLICE, key)

    #---------------------------------------------------------------------------
    def __add__(self, other: tp.Any) -> tp.Any:
        return self._container._ufunc_binary_operator(
                operator=OPERATORS['__add__'],
                other=other,
                axis=self._axis,
                fill_value=self._fill_value,
                )

    def __sub__(self, other: tp.Any) -> tp.Any:
        return self._container._ufunc_binary_operator(
                operator=OPERATORS['__sub__'],
                other=other,
                axis=self._axis,
                fill_value=self._fill_value,
                )

    def __mul__(self, other: tp.Any) -> tp.Any:
        return self._container._ufunc_binary_operator(
                operator=OPERATORS['__mul__'],
                other=other,
                axis=self._axis,
                fill_value=self._fill_value,
                )

    # def __matmul__(self, other: tp.Any) -> tp.Any:
    #     return self._container._ufunc_binary_operator(
    #             operator=OPERATORS['__matmul__'],
    #             other=other,
    #             axis=self._axis,
    #             fill_value=self._fill_value,
    #             )

    def __truediv__(self, other: tp.Any) -> tp.Any:
        return self._container._ufunc_binary_operator(
                operator=OPERATORS['__truediv__'],
                other=other,
                axis=self._axis,
                fill_value=self._fill_value,
                )

    def __floordiv__(self, other: tp.Any) -> tp.Any:
        return self._container._ufunc_binary_operator(
                operator=OPERATORS['__floordiv__'],
                other=other,
                axis=self._axis,
                fill_value=self._fill_value,
                )

    def __mod__(self, other: tp.Any) -> tp.Any:
        return self._container._ufunc_binary_operator(
                operator=OPERATORS['__mod__'],
                other=other,
                axis=self._axis,
                fill_value=self._fill_value,
                )

    def __pow__(self, other: tp.Any) -> tp.Any:
        return self._container._ufunc_binary_operator(
                operator=OPERATORS['__pow__'],
                other=other,
                axis=self._axis,
                fill_value=self._fill_value,
                )

    def __lshift__(self, other: tp.Any) -> tp.Any:
        return self._container._ufunc_binary_operator(
                operator=OPERATORS['__lshift__'],
                other=other,
                axis=self._axis,
                fill_value=self._fill_value,
                )

    def __rshift__(self, other: tp.Any) -> tp.Any:
        return self._container._ufunc_binary_operator(
                operator=OPERATORS['__rshift__'],
                other=other,
                axis=self._axis,
                fill_value=self._fill_value,
                )

    def __and__(self, other: tp.Any) -> tp.Any:
        return self._container._ufunc_binary_operator(
                operator=OPERATORS['__and__'],
                other=other,
                axis=self._axis,
                fill_value=self._fill_value,
                )

    def __xor__(self, other: tp.Any) -> tp.Any:
        return self._container._ufunc_binary_operator(
                operator=OPERATORS['__xor__'],
                other=other,
                axis=self._axis,
                fill_value=self._fill_value,
                )

    def __or__(self, other: tp.Any) -> tp.Any:
        return self._container._ufunc_binary_operator(
                operator=OPERATORS['__or__'],
                other=other,
                axis=self._axis,
                fill_value=self._fill_value,
                )

    def __lt__(self, other: tp.Any) -> tp.Any:
        return self._container._ufunc_binary_operator(
                operator=OPERATORS['__lt__'],
                other=other,
                axis=self._axis,
                fill_value=self._fill_value,
                )

    def __le__(self, other: tp.Any) -> tp.Any:
        return self._container._ufunc_binary_operator(
                operator=OPERATORS['__le__'],
                other=other,
                axis=self._axis,
                fill_value=self._fill_value,
                )

    def __eq__(self, other: tp.Any) -> tp.Any:
        return self._container._ufunc_binary_operator(
                operator=OPERATORS['__eq__'],
                other=other,
                axis=self._axis,
                fill_value=self._fill_value,
                )

    def __ne__(self, other: tp.Any) -> tp.Any:
        return self._container._ufunc_binary_operator(
                operator=OPERATORS['__ne__'],
                other=other,
                axis=self._axis,
                fill_value=self._fill_value,
                )

    def __gt__(self, other: tp.Any) -> tp.Any:
        return self._container._ufunc_binary_operator(
                operator=OPERATORS['__gt__'],
                other=other,
                axis=self._axis,
                fill_value=self._fill_value,
                )

    def __ge__(self, other: tp.Any) -> tp.Any:
        return self._container._ufunc_binary_operator(
                operator=OPERATORS['__ge__'],
                other=other,
                axis=self._axis,
                fill_value=self._fill_value,
                )

    #---------------------------------------------------------------------------
    def __radd__(self, other: tp.Any) -> tp.Any:
        return self._container._ufunc_binary_operator(
                operator=OPERATORS['__radd__'],
                other=other,
                axis=self._axis,
                fill_value=self._fill_value,
                )

    def __rsub__(self, other: tp.Any) -> tp.Any:
        return self._container._ufunc_binary_operator(
                operator=OPERATORS['__rsub__'],
                other=other,
                axis=self._axis,
                fill_value=self._fill_value,
                )

    def __rmul__(self, other: tp.Any) -> tp.Any:
        return self._container._ufunc_binary_operator(
                operator=OPERATORS['__rmul__'],
                other=other,
                axis=self._axis,
                fill_value=self._fill_value,
                )

    # def __rmatmul__(self, other: tp.Any) -> tp.Any:
    #     return self._container._ufunc_binary_operator(
    #             operator=OPERATORS['__rmatmul__'],
    #             other=other,
    #             axis=self._axis,
    #             fill_value=self._fill_value,
    #             )

    def __rtruediv__(self, other: tp.Any) -> tp.Any:
        return self._container._ufunc_binary_operator(
                operator=OPERATORS['__rtruediv__'],
                other=other,
                axis=self._axis,
                fill_value=self._fill_value,
                )

    def __rfloordiv__(self, other: tp.Any) -> tp.Any:
        return self._container._ufunc_binary_operator(
                operator=OPERATORS['__rfloordiv__'],
                other=other,
                axis=self._axis,
                fill_value=self._fill_value,
                )

#---------------------------------------------------------------------------
class InterfaceFillValueGO(InterfaceFillValue[TContainer]): # only type is FrameGO

    __slots__ = ()
    INTERFACE = InterfaceFillValue.INTERFACE + ( #type: ignore
            '__setitem__',
            )

    def __setitem__(self,
            key: tp.Hashable,
            value: tp.Any,
            ) -> None:
        self._container.__setitem__(key, value, self._fill_value) #type: ignore

#---------------------------------------------------------------------------


class InterfaceBatchFillValue(InterfaceBatch):
    '''Alternate string interface specialized for the :obj:`Batch`.
    '''
    INTERFACE = INTERFACE_FILL_VALUE

    __slots__ = (
            '_batch_apply',
            '_fill_value',
            '_axis',
            )

    def __init__(self,
            batch_apply: tp.Callable[[AnyCallable], 'Batch'],
            fill_value: object = np.nan,
            axis: int = 0,
            ) -> None:

        self._batch_apply = batch_apply
        self._fill_value = fill_value
        self._axis = axis


    #---------------------------------------------------------------------------
    @property
    def via_T(self) -> 'InterfaceBatchTranspose':
        '''
        Interface for using binary operators with one-dimensional sequences, where the opperand is applied column-wise.
        '''
        from static_frame.core.node_transpose import InterfaceBatchTranspose
        return InterfaceBatchTranspose(
                batch_apply=self._batch_apply,
                fill_value=self._fill_value,
                )

    #---------------------------------------------------------------------------
    @property
    def loc(self) -> InterfaceGetItem['Frame']:
        '''Label-based selection where labels not specified will define a new container containing those labels filled with the fill value.
        '''
        def func(key: GetItemKeyType) -> 'Batch':
            return self._batch_apply(
                    lambda c: InterfaceFillValue(c,
                            fill_value=self._fill_value,
                            axis=self._axis).loc[key]
                    )
        return InterfaceGetItem(func)

    def __getitem__(self,  key: GetItemKeyType) -> 'Batch':
        '''Label-based selection where labels not specified will define a new container containing those labels filled with the fill value.
        '''
        return self._batch_apply(
                lambda c: InterfaceFillValue(c,
                        fill_value=self._fill_value,
                        axis=self._axis).__getitem__(key)
                )

    #---------------------------------------------------------------------------
    def __add__(self, other: tp.Any) -> 'Batch':
        return self._batch_apply(
            lambda c: InterfaceFillValue(c,
                    fill_value=self._fill_value,
                    axis=self._axis).__add__(other)
            )

    def __sub__(self, other: tp.Any) -> 'Batch':
        return self._batch_apply(
            lambda c: InterfaceFillValue(c,
                    fill_value=self._fill_value,
                    axis=self._axis).__sub__(other)
            )

    def __mul__(self, other: tp.Any) -> 'Batch':
        return self._batch_apply(
            lambda c: InterfaceFillValue(c,
                    fill_value=self._fill_value,
                    axis=self._axis).__mul__(other)
            )

    # def __matmul__(self, other: tp.Any) -> 'Batch':

    def __truediv__(self, other: tp.Any) -> 'Batch':
        return self._batch_apply(
            lambda c: InterfaceFillValue(c,
                    fill_value=self._fill_value,
                    axis=self._axis).__truediv__(other)
            )

    def __floordiv__(self, other: tp.Any) -> 'Batch':
        return self._batch_apply(
            lambda c: InterfaceFillValue(c,
                    fill_value=self._fill_value,
                    axis=self._axis).__floordiv__(other)
            )

    def __mod__(self, other: tp.Any) -> 'Batch':
        return self._batch_apply(
            lambda c: InterfaceFillValue(c,
                    fill_value=self._fill_value,
                    axis=self._axis).__mod__(other)
            )

    def __pow__(self, other: tp.Any) -> 'Batch':
        return self._batch_apply(
            lambda c: InterfaceFillValue(c,
                    fill_value=self._fill_value,
                    axis=self._axis).__pow__(other)
            )

    def __lshift__(self, other: tp.Any) -> 'Batch':
        return self._batch_apply(
            lambda c: InterfaceFillValue(c,
                    fill_value=self._fill_value,
                    axis=self._axis).__lshift__(other)
            )

    def __rshift__(self, other: tp.Any) -> 'Batch':
        return self._batch_apply(
            lambda c: InterfaceFillValue(c,
                    fill_value=self._fill_value,
                    axis=self._axis).__rshift__(other)
            )

    def __and__(self, other: tp.Any) -> 'Batch':
        return self._batch_apply(
            lambda c: InterfaceFillValue(c,
                    fill_value=self._fill_value,
                    axis=self._axis).__and__(other)
            )

    def __xor__(self, other: tp.Any) -> 'Batch':
        return self._batch_apply(
            lambda c: InterfaceFillValue(c,
                    fill_value=self._fill_value,
                    axis=self._axis).__xor__(other)
            )

    def __or__(self, other: tp.Any) -> 'Batch':
        return self._batch_apply(
            lambda c: InterfaceFillValue(c,
                    fill_value=self._fill_value,
                    axis=self._axis).__or__(other)
            )

    def __lt__(self, other: tp.Any) -> 'Batch':
        return self._batch_apply(
            lambda c: InterfaceFillValue(c,
                    fill_value=self._fill_value,
                    axis=self._axis).__lt__(other)
            )

    def __le__(self, other: tp.Any) -> 'Batch':
        return self._batch_apply(
            lambda c: InterfaceFillValue(c,
                    fill_value=self._fill_value,
                    axis=self._axis).__le__(other)
            )

    def __eq__(self, other: tp.Any) -> 'Batch': #type: ignore
        return self._batch_apply(
            lambda c: InterfaceFillValue(c,
                    fill_value=self._fill_value,
                    axis=self._axis).__eq__(other)
            )

    def __ne__(self, other: tp.Any) -> 'Batch': #type: ignore
        return self._batch_apply(
            lambda c: InterfaceFillValue(c,
                    fill_value=self._fill_value,
                    axis=self._axis).__ne__(other)
            )

    def __gt__(self, other: tp.Any) -> 'Batch':
        return self._batch_apply(
            lambda c: InterfaceFillValue(c,
                    fill_value=self._fill_value,
                    axis=self._axis).__gt__(other)
            )

    def __ge__(self, other: tp.Any) -> 'Batch':
        return self._batch_apply(
            lambda c: InterfaceFillValue(c,
                    fill_value=self._fill_value,
                    axis=self._axis).__ge__(other)
            )

    #---------------------------------------------------------------------------
    def __radd__(self, other: tp.Any) -> 'Batch':
        return self._batch_apply(
            lambda c: InterfaceFillValue(c,
                    fill_value=self._fill_value,
                    axis=self._axis).__radd__(other)
            )

    def __rsub__(self, other: tp.Any) -> 'Batch':
        return self._batch_apply(
            lambda c: InterfaceFillValue(c,
                    fill_value=self._fill_value,
                    axis=self._axis).__rsub__(other)
            )

    def __rmul__(self, other: tp.Any) -> 'Batch':
        return self._batch_apply(
            lambda c: InterfaceFillValue(c,
                    fill_value=self._fill_value,
                    axis=self._axis).__rmul__(other)
            )

    # def __rmatmul__(self, other: tp.Any) -> 'Batch':

    def __rtruediv__(self, other: tp.Any) -> 'Batch':
        return self._batch_apply(
            lambda c: InterfaceFillValue(c,
                    fill_value=self._fill_value,
                    axis=self._axis).__rtruediv__(other)
            )

    def __rfloordiv__(self, other: tp.Any) -> 'Batch':
        return self._batch_apply(
            lambda c: InterfaceFillValue(c,
                    fill_value=self._fill_value,
                    axis=self._axis).__rfloordiv__(other)
            )



import gzip
import io
# import json
import os
import typing as tp
import unittest
from pathlib import Path
from unittest.mock import MagicMock
from unittest.mock import patch
from zipfile import ZipFile

from static_frame.core.frame import Frame
from static_frame.core.www import WWW
from static_frame.core.www import BytesIOTemporaryFile
from static_frame.core.www import StringIOTemporaryFile
from static_frame.test.test_case import TestCase
from static_frame.test.test_case import temp_file

URL = 'http://foo'

def prepare_mock(mock: MagicMock, content: tp.Union[str, bytes]) -> None:
    if isinstance(content, str):
        payload = io.BytesIO(bytes(content, encoding='utf-8'))
    else:
        payload = io.BytesIO(content)
    cm = MagicMock()
    cm.__enter__.return_value.read = payload.read
    mock.return_value = cm

def load_zip_in_bytes(content: tp.Dict[str, tp.Union[str, bytes]]) -> io.BytesIO:
    archive = io.BytesIO()
    with ZipFile(archive, mode='w') as zf:
        for label, value in content.items():
            if isinstance(value, bytes):
                zf.writestr(label, value)
            else:
                zf.writestr(label, value.encode('utf-8'))
    archive.seek(0)
    return archive

def load_gzip_in_bytes(content: tp.Union[str, bytes]) -> io.BytesIO:
    archive = io.BytesIO()
    with gzip.open(archive, mode='wb') as gz:
        if isinstance(content, bytes):
            gz.write(content)
        else:
            gz.write(content.encode('utf-8'))
    archive.seek(0)
    return archive



class TestUnit(TestCase):

    def test_www_from_file_a(self) -> None:
        with patch('urllib.request.urlopen') as mock:
            prepare_mock(mock, 'foo')

            post = WWW.from_file(URL, encoding='utf-8', in_memory=True)
            self.assertTrue(isinstance(post, io.StringIO))
            self.assertEqual(post.read(), 'foo')

    def test_www_from_file_b(self) -> None:
        with patch('urllib.request.urlopen') as mock:
            prepare_mock(mock, 'bar')

            post = WWW.from_file(URL, encoding='utf-8', in_memory=False)
            self.assertTrue(isinstance(post, StringIOTemporaryFile))
            self.assertEqual('bar', post.read())

    def test_www_from_file_c(self) -> None:
        with patch('urllib.request.urlopen') as mock:
            prepare_mock(mock, 'bar')

            post = WWW.from_file(URL, encoding=None, in_memory=True)
            self.assertTrue(isinstance(post, io.BytesIO))
            self.assertEqual(post.read(), b'bar')

    def test_www_from_file_d(self) -> None:
        with patch('urllib.request.urlopen') as mock:
            prepare_mock(mock, 'foo')

            post = WWW.from_file(URL, encoding=None, in_memory=False)

            self.assertTrue(isinstance(post, BytesIOTemporaryFile))
            self.assertEqual(b'foo', post.read())


    def test_www_from_file_from_delimited_a(self) -> None:
        with patch('urllib.request.urlopen') as mock:
            prepare_mock(mock, 'a,b,c\n1,True,x\n20,False,y\n')

            post = Frame.from_csv(WWW.from_file(URL))
            self.assertEqual(post.to_pairs(),
                    (('a', ((0, 1), (1, 20))), ('b', ((0, True), (1, False))), ('c', ((0, 'x'), (1, 'y')))))

    # def test_url_from_delimited_b(self) -> None:1

    #     url = 'https://stats.govt.nz/assets/Uploads/Business-financial-data/Business-financial-data-June-2022-quarter/Download-data/business-financial-data-june-2022-quarter-csv.zip'

    #     post = URL(url)
    #     import ipdb; ipdb.set_trace()

    #---------------------------------------------------------------------------

    def test_string_io_temp_file_a(self) -> None:
        content = 'foo\nbar'
        with temp_file('.txt') as fp:
            with open(fp, 'w', encoding='utf-8') as f:
                f.write(content)

            siotf = StringIOTemporaryFile(fp, encoding='utf-8')

            self.assertEqual(siotf.read(), content)
            siotf.seek(0)

            self.assertEqual(siotf.readline(), 'foo\n')
            siotf.seek(0)

            self.assertEqual(tuple(siotf), ('foo\n', 'bar'))

            del siotf
            self.assertFalse(os.path.exists(fp))

            # restore file so context manager can clean up
            with open(fp, 'w', encoding='utf-8') as f:
                f.write(content)

    def test_bytes_io_temp_file_a(self) -> None:
        content = b'foo\nbar'

        with temp_file('.txt') as fp:
            with open(fp, 'wb') as f:
                f.write(content)

            siotf = BytesIOTemporaryFile(fp)

            self.assertEqual(siotf.read(), content)
            siotf.seek(0)

            self.assertEqual(siotf.readline(), b'foo\n')
            siotf.seek(0)

            self.assertEqual(tuple(siotf), (b'foo\n', b'bar'))

            del siotf
            self.assertFalse(os.path.exists(fp))

            # restore file so context manager can clean up
            with open(fp, 'wb') as f:
                f.write(content)

    #---------------------------------------------------------------------------

    def test_download_archive_a(self) -> None:
        archive = load_zip_in_bytes({'foo': 'a,b,c\n1,True,x\n2,False,y\n'})

        with patch('urllib.request.urlopen') as mock:
            prepare_mock(mock, archive.read())

            post = WWW._download_archive('http://foo',
                    in_memory=True,
                    buffer_size=1024,
                    extension='.zip')

            archive.seek(0)
            assert isinstance(post, io.BytesIO)
            self.assertEqual(post.read(), archive.read())

    def test_download_archive_b(self) -> None:
        content = b'a,b,c\n1,True,x\n2,False,y\n'
        archive = load_zip_in_bytes({'foo': content})

        with patch('urllib.request.urlopen') as mock:
            prepare_mock(mock, archive.read())

            post = WWW._download_archive(URL,
                    in_memory=False,
                    buffer_size=1024,
                    extension='.zip')
            self.assertIsInstance(post, Path)

            with ZipFile(post) as zf:
                contained = zf.read('foo')
                self.assertEqual(contained, content)

    #---------------------------------------------------------------------------
    def test_www_from_zip_a(self) -> None:
        content = b'a,b,c\n1,True,x\n2,False,y\n'
        archive = load_zip_in_bytes({'foo': content, 'bar': content})

        with patch('urllib.request.urlopen') as mock:
            prepare_mock(mock, archive.read())
            with self.assertRaises(RuntimeError):
                post = WWW.from_zip(URL)

    def test_www_from_zip_b(self) -> None:
        content1 = 'a,b,c\n1,True,x\n2,False,y\n'
        content2 = 'p,q\nTrue,x\nFalse,y\n'
        archive = load_zip_in_bytes({'foo': content1, 'bar': content2})

        with patch('urllib.request.urlopen') as mock:
            prepare_mock(mock, archive.read())
            post = WWW.from_zip(URL, component='foo')
            self.assertEqual(post.read(), content1)

    def test_www_from_zip_c(self) -> None:
        content2 = 'p,q\nTrue,x\nFalse,y\n'
        archive = load_zip_in_bytes({'bar': content2})

        with patch('urllib.request.urlopen') as mock:
            prepare_mock(mock, archive.read())
            post = WWW.from_zip(URL)
            self.assertEqual(post.read(), content2)

    def test_www_from_zip_d(self) -> None:
        content2 = b'p,q\nTrue,x\nFalse,y\n'
        archive = load_zip_in_bytes({'bar': content2})

        with patch('urllib.request.urlopen') as mock:
            prepare_mock(mock, archive.read())
            post = WWW.from_zip(URL, encoding=None)
            self.assertEqual(post.read(), content2)

    def test_www_from_zip_e(self) -> None:
        content2 = 'p,q\nTrue,x\nFalse,y\n'
        archive = load_zip_in_bytes({'bar': content2})

        with patch('urllib.request.urlopen') as mock:
            prepare_mock(mock, archive.read())

            with temp_file('.txt') as fp:
                with self.assertRaises(RuntimeError):
                    _ = WWW.from_zip(URL, fp=fp, in_memory=True)

                post = WWW.from_zip(URL, fp=fp)
                self.assertEqual(str(post), fp)

                with open(post, encoding='utf-8') as postf:
                    self.assertEqual(postf.read(), content2)

    #---------------------------------------------------------------------------
    def test_www_from_gzip_a(self) -> None:
        content = 'p,q\nTrue,x\nFalse,y\n'
        archive = load_gzip_in_bytes(content)

        with patch('urllib.request.urlopen') as mock:
            prepare_mock(mock, archive.read())
            post = WWW.from_gzip(URL)
            self.assertEqual(post.read(), content)

    def test_www_from_gzip_b(self) -> None:
        content = b'p,q\nTrue,x\nFalse,y\n'
        archive = load_gzip_in_bytes(content)

        with patch('urllib.request.urlopen') as mock:
            prepare_mock(mock, archive.read())
            post = WWW.from_gzip(URL, encoding=None)
            self.assertEqual(post.read(), content)

    def test_www_from_gzip_c(self) -> None:
        content = 'p,q\nTrue,x\nFalse,y\n'
        archive = load_gzip_in_bytes(content)

        with patch('urllib.request.urlopen') as mock:
            prepare_mock(mock, archive.read())

            with temp_file('.txt') as fp:
                post = WWW.from_gzip(URL, fp=fp)
                self.assertEqual(str(post), fp)

                with open(post, encoding='utf-8') as postf:
                    self.assertEqual(postf.read(), content)

    #---------------------------------------------------------------------------
    def test_frame_from_json_a(self) -> None:

        content = '''[
        {
        "userId": 1,
        "id": 1,
        "title": "delectus aut autem",
        "completed": false
        },
        {
        "userId": 1,
        "id": 2,
        "title": "quis ut nam facilis et officia qui",
        "completed": false
        }]'''
        with patch('urllib.request.urlopen') as mock:
            prepare_mock(mock, content)
            post = Frame.from_json_url(url=URL)
            self.assertEqual(post.shape, (2, 4))
            self.assertEqual([dt.kind for dt in post.dtypes.values], ['i', 'i', 'U', 'b'])

if __name__ == '__main__':
    unittest.main()



import typing as tp

import numpy as np
from arraykit import column_2d_filter

from static_frame.core.node_selector import Interface
from static_frame.core.node_selector import InterfaceBatch
from static_frame.core.type_blocks import TypeBlocks
from static_frame.core.util import AnyCallable
from static_frame.core.util import DtypeSpecifier
from static_frame.core.util import UFunc
from static_frame.core.util import blocks_to_array_2d

if tp.TYPE_CHECKING:
    from static_frame.core.batch import Batch  # pylint: disable = W0611 #pragma: no cover
    from static_frame.core.frame import Frame  # pylint: disable = W0611 #pragma: no cover
    from static_frame.core.index import Index  # pylint: disable = W0611 #pragma: no cover
    from static_frame.core.index_hierarchy import IndexHierarchy  # pylint: disable = W0611 #pragma: no cover
    from static_frame.core.series import Series  # pylint: disable = W0611 #pragma: no cover

TContainer = tp.TypeVar('TContainer',
        'Frame',
        'IndexHierarchy',
        'Series',
        'Index',
        )

INTERFACE_VALUES = (
        'apply',
        '__array_ufunc__',
        '__call__',
        )

VALID_UFUNC_ARRAY_METHODS = frozenset(('__call__',))


class InterfaceValues(Interface[TContainer]):
    '''
    If a user wants to call a ufunc and get back an array of variable dimensionality, they have to call that ufunc on one consolidated array via .values; any attempt at block-level manipulation will have to, under some scenarios, figure out how to combine the per-block results (and an appropriate type) into an array. This is undesirable. Instead, all applications of this interface must use UFuncs that retain dimensionality.
    '''
    __slots__ = (
            '_container',
            '_consolidate_blocks',
            '_unify_blocks',
            '_dtype',
            )
    INTERFACE = INTERFACE_VALUES

    def __init__(self,
            container: TContainer,
            *,
            consolidate_blocks: bool = False,
            unify_blocks: bool = False,
            dtype: DtypeSpecifier = None,
            ) -> None:
        self._container: TContainer = container
        self._consolidate_blocks = consolidate_blocks
        self._unify_blocks = unify_blocks
        self._dtype = dtype

    def __call__(self,
            *,
            consolidate_blocks: bool = False,
            unify_blocks: bool = False,
            dtype: DtypeSpecifier = None,
            ) -> 'InterfaceValues[TContainer]':
        '''
        Args:
            consolidate_blocks: Group adjacent same-typed arrays into 2D arrays.
            unify_blocks: Group all arrays into single array, re-typing to an appropriate dtype.
            dtype: specify a dtype to be used in conversion before consolidation or unification, and before function application.
        '''
        return self.__class__(self._container,
                consolidate_blocks=consolidate_blocks,
                unify_blocks=unify_blocks,
                dtype=dtype,
                )

    def __array_ufunc__(self,
            ufunc: UFunc,
            method: str,
            *args: tp.Any,
            **kwargs: tp.Any,
            ) -> TContainer:
        '''Support for applying NumPy functions directly on containers, returning NumPy arrays.
        '''
        from static_frame.core.frame import Frame
        from static_frame.core.series import Series


        if method not in VALID_UFUNC_ARRAY_METHODS:
            return NotImplemented #pragma: no cover

        def func(block: np.ndarray, normalize_2d: bool = True) -> np.ndarray:
            if normalize_2d:
                block = column_2d_filter(block)

            # NOTE: we assume that our target array (the passed in block) should alwasy be the first argument; then, we filter out arguments that are either this object or an InterfaceBatchValues instance
            args_final = [block]
            for arg in args:
                if arg is self or isinstance(arg, InterfaceBatchValues):
                    continue
                args_final.append(arg)
            # [(arg if arg is not self else block) for arg in args]
            array = ufunc(*args_final, **kwargs)
            array.flags.writeable = False
            return array

        if self._container._NDIM == 2:
            blocks: tp.Iterable[np.ndarray] = self._container._blocks._blocks #type: ignore

            if self._unify_blocks:
                dtype = self._container._blocks._row_dtype if self._dtype is None else self._dtype #type: ignore
                tb = TypeBlocks.from_blocks(func(blocks_to_array_2d(
                        blocks=blocks,
                        shape=self._container.shape,
                        dtype=dtype,
                        )))
            elif self._consolidate_blocks:
                if self._dtype is not None:
                    blocks = (b.astype(self._dtype) for b in blocks)
                tb = TypeBlocks.from_blocks(
                        func(b) for b in TypeBlocks.consolidate_blocks(blocks)
                        )
            else:
                if self._dtype is not None:
                    blocks = (func(b.astype(self._dtype)) for b in blocks)
                else:
                    blocks = (func(b) for b in blocks)
                tb = TypeBlocks.from_blocks(blocks)

            if isinstance(self._container, Frame):
                return self._container.__class__(
                        tb,
                        index=self._container.index,
                        columns=self._container.columns,
                        name=self._container.name,
                        own_index=True,
                        own_data=True,
                        own_columns=self._container.STATIC,
                        )
            #IndexHierarchy
            return self._container._from_type_blocks( #type: ignore
                tb,
                index_constructors=self._container._index_constructors, # type: ignore
                name=self._container._name,
                own_blocks=True,
                )
        # all 1D containers
        if self._dtype is not None:
            values = func(self._container.values.astype(self._dtype),
                    normalize_2d=False,
                    )
        else:
            values = func(self._container.values,
                    normalize_2d=False,
                    )

        if isinstance(self._container, Series):
            return self._container.__class__(values,
                    index=self._container.index,
                    name=self._container.name,
                    own_index=True,
                    )
        # else, Index
        return self._container.__class__(values,
                name=self._container.name,
                )


    def apply(self,
            func: UFunc,
            *args: tp.Any,
            **kwargs: tp.Any,
            ) -> TContainer:
        return self.__array_ufunc__(
                func,
                '__call__',
                *args,
                **kwargs,
                )

class InterfaceBatchValues(InterfaceBatch):

    __slots__ = (
            '_batch_apply',
            '_consolidate_blocks',
            '_unify_blocks',
            '_dtype',
            )
    INTERFACE = INTERFACE_VALUES

    def __init__(self,
            batch_apply: tp.Callable[[AnyCallable], 'Batch'],
            *,
            consolidate_blocks: bool = False,
            unify_blocks: bool = False,
            dtype: DtypeSpecifier = None,
            ) -> None:
        self._batch_apply = batch_apply
        self._consolidate_blocks = consolidate_blocks
        self._unify_blocks = unify_blocks
        self._dtype = dtype

    #---------------------------------------------------------------------------
    def apply(self,
            func: UFunc,
            *args: tp.Any,
            **kwargs: tp.Any,
            ) -> 'Batch':
        '''
        Interface for using binary operators and methods with a pre-defined fill value.
        '''
        return self._batch_apply(lambda c: c.via_values(
                consolidate_blocks=self._consolidate_blocks,
                unify_blocks=self._unify_blocks,
                dtype=self._dtype,
                ).apply(func, *args, **kwargs))

    def __call__(self,
            *,
            consolidate_blocks: bool = False,
            unify_blocks: bool = False,
            dtype: DtypeSpecifier = None,
            ) -> 'InterfaceBatchValues':
        '''
        Args:
            consolidate_blocks: Group adjacent same-typed arrays into 2D arrays.
            unify_blocks: Group all arrays into single array, re-typing to an appropriate dtype.
            dtype: specify a dtype to be used in conversion before consolidation or unification, and before function application.
        '''
        return self.__class__(self._batch_apply,
                consolidate_blocks=consolidate_blocks,
                unify_blocks=unify_blocks,
                dtype=dtype,
                )

    def __array_ufunc__(self,
            ufunc: UFunc,
            method: str,
            *args: tp.Any,
            **kwargs: tp.Any,
            ) -> 'Batch':
        '''Support for applying NumPy functions directly on containers, returning NumPy arrays.
        '''
        # NOTE: want to fail method is not supported at call time of this function, not the deferred execution via Batch
        if method not in VALID_UFUNC_ARRAY_METHODS:
            return NotImplemented #pragma: no cover

        def func(c: TContainer) -> np.ndarray:
            return c.via_values(
                    consolidate_blocks=self._consolidate_blocks,
                    unify_blocks=self._unify_blocks,
                    dtype=self._dtype,
                    ).__array_ufunc__(ufunc,
                            method,
                            *args,
                            **kwargs,
                            )

        return self._batch_apply(func)




'''
Tools for iterators in Series and Frame. These components are imported by both series.py and frame.py; these components also need to be able to return Series and Frame, and thus use deferred, function-based imports.
'''

import typing as tp
from concurrent.futures import ProcessPoolExecutor
from concurrent.futures import ThreadPoolExecutor
from enum import Enum
from functools import partial

import numpy as np
from arraykit import name_filter

from static_frame.core.container_util import group_from_container
from static_frame.core.doc_str import doc_inject
from static_frame.core.util import KEY_ITERABLE_TYPES
from static_frame.core.util import AnyCallable
from static_frame.core.util import DepthLevelSpecifier
from static_frame.core.util import DtypeSpecifier
from static_frame.core.util import IndexConstructor
from static_frame.core.util import Mapping
from static_frame.core.util import NameType
from static_frame.core.util import TupleConstructorType
from static_frame.core.util import iterable_to_array_1d

# import multiprocessing as mp
# mp_context = mp.get_context('spawn')


# from static_frame.core.util import array_from_iterator


if tp.TYPE_CHECKING:
    from static_frame.core.bus import Bus  # pylint: disable=W0611 #pragma: no cover
    from static_frame.core.frame import Frame  # pylint: disable=W0611 #pragma: no cover
    from static_frame.core.index import Index  # pylint: disable=W0611 #pragma: no cover
    from static_frame.core.quilt import Quilt  # pylint: disable=W0611 #pragma: no cover
    from static_frame.core.series import Series  # pylint: disable=W0611 #pragma: no cover
    from static_frame.core.yarn import Yarn  # pylint: disable=W0611 #pragma: no cover


FrameOrSeries = tp.TypeVar('FrameOrSeries', 'Frame', 'Series', 'Bus', 'Quilt', 'Yarn')
PoolArgGen = tp.Callable[[], tp.Union[tp.Iterator[tp.Any], tp.Iterator[tp.Tuple[tp.Any, tp.Any]]]]
# FrameSeriesIndex = tp.TypeVar('FrameSeriesIndex', 'Frame', 'Series', 'Index')


class IterNodeType(Enum):
    VALUES = 1
    ITEMS = 2


class IterNodeApplyType(Enum):
    SERIES_VALUES = 0
    SERIES_ITEMS = 1 # only used for iter_window_*
    SERIES_ITEMS_GROUP_VALUES = 2
    SERIES_ITEMS_GROUP_LABELS = 3
    FRAME_ELEMENTS = 4
    INDEX_LABELS = 5

    @classmethod
    def is_items(cls, apply_type: 'IterNodeApplyType') -> bool:
        '''Return True if the apply_constructor to be used consumes items; otherwise, the apply_constructor consumes values alone.
        '''
        if (apply_type is cls.SERIES_VALUES
                or apply_type is cls.INDEX_LABELS
                ):
            return False
        return True



class IterNodeDelegate(tp.Generic[FrameOrSeries]):
    '''
    Delegate returned from :obj:`static_frame.IterNode`, providing iteration as well as a family of apply methods.
    '''

    __slots__ = (
            '_func_values',
            '_func_items',
            '_yield_type',
            '_apply_constructor',
            '_apply_type',
            )

    INTERFACE: tp.Tuple[str, ...] = (
            'apply',
            'apply_iter',
            'apply_iter_items',
            'apply_pool',
            )

    def __init__(self,
            func_values: tp.Callable[..., tp.Iterable[tp.Any]],
            func_items: tp.Callable[..., tp.Iterable[tp.Tuple[tp.Any, tp.Any]]],
            yield_type: IterNodeType,
            apply_constructor: tp.Callable[..., FrameOrSeries],
            apply_type: IterNodeApplyType,
        ) -> None:
        '''
        Args:
            apply_constructor: Callable (generally a class) used to construct the object returned from apply(); must take an iterator of items.
        '''
        self._func_values = func_values
        self._func_items = func_items
        self._yield_type = yield_type
        self._apply_constructor: tp.Callable[..., FrameOrSeries] = apply_constructor
        self._apply_type = apply_type

    #---------------------------------------------------------------------------

    def _apply_iter_items_parallel(self,
            func: AnyCallable,
            max_workers: tp.Optional[int] = None,
            chunksize: int = 1,
            use_threads: bool = False,
            ) -> tp.Iterator[tp.Tuple[tp.Any, tp.Any]]:

        pool_executor = ThreadPoolExecutor if use_threads else ProcessPoolExecutor

        if not callable(func): # support array, Series mapping
            func = getattr(func, '__getitem__')

        # use side effect list population to create keys when iterating over values
        func_keys = []
        arg_gen: PoolArgGen

        if self._yield_type is IterNodeType.VALUES:
            def arg_gen() -> tp.Iterator[tp.Any]: #pylint: disable=E0102
                for k, v in self._func_items():
                    func_keys.append(k)
                    yield v
        else:
            def arg_gen() -> tp.Iterator[tp.Tuple[tp.Any, tp.Any]]: #pylint: disable=E0102
                for k, v in self._func_items():
                    func_keys.append(k)
                    yield k, v

        with pool_executor(max_workers=max_workers) as executor:
            yield from zip(func_keys,
                    executor.map(func, arg_gen(), chunksize=chunksize)
                    )

    def _apply_iter_parallel(self,
            func: AnyCallable,
            max_workers: tp.Optional[int] = None,
            chunksize: int = 1,
            use_threads: bool = False,
            ) -> tp.Iterator[tp.Any]:

        pool_executor = ThreadPoolExecutor if use_threads else ProcessPoolExecutor
        if not callable(func): # support array, Series mapping
            func = getattr(func, '__getitem__')

        # use side effect list population to create keys when iterating over values
        arg_gen = (self._func_values if self._yield_type is IterNodeType.VALUES
                else self._func_items)

        with pool_executor(max_workers=max_workers) as executor:
            yield from executor.map(func, arg_gen(), chunksize=chunksize)

    #---------------------------------------------------------------------------
    @doc_inject(selector='apply')
    def apply_iter_items(self,
            func: AnyCallable,
            ) -> tp.Iterator[tp.Tuple[tp.Any, tp.Any]]:
        '''
        {doc} A generator of resulting key, value pairs.

        Args:
            {func}

        Yields:
            Pairs of label, value after function application.
        '''
        # depend on yield type, we determine what the passed in function expects to
        if self._yield_type is IterNodeType.VALUES:
            yield from ((k, func(v)) for k, v in self._func_items())
        else:
            yield from ((k, func(k, v)) for k, v in self._func_items())

    @doc_inject(selector='apply')
    def apply_iter(self,
            func: AnyCallable
            ) -> tp.Iterator[tp.Any]:
        '''
        {doc} A generator of resulting values.

        Args:
            {func}

        Yields:
            Values after function application.
        '''
        if self._yield_type is IterNodeType.VALUES:
            yield from (func(v) for v in self._func_values())
        else:
            yield from (func(k, v) for k, v in self._func_items())

    @doc_inject(selector='apply')
    def apply(self,
            func: AnyCallable,
            *,
            dtype: DtypeSpecifier = None,
            name: NameType = None,
            index_constructor: tp.Optional[IndexConstructor]= None,
            ) -> FrameOrSeries:
        '''
        {doc} Returns a new container.

        Args:
            {func}
            {dtype}
        '''
        if not callable(func):
            raise RuntimeError('use map_fill(), map_any(), or map_all() for applying a mapping type')

        if IterNodeApplyType.is_items(self._apply_type):
            apply_func = self.apply_iter_items
        else:
            apply_func = self.apply_iter

        return self._apply_constructor(
                apply_func(func),
                dtype=dtype,
                name=name,
                index_constructor=index_constructor,
                )

    @doc_inject(selector='apply')
    def apply_pool(self,
            func: AnyCallable,
            *,
            dtype: DtypeSpecifier = None,
            name: NameType = None,
            index_constructor: tp.Optional[IndexConstructor]= None,
            max_workers: tp.Optional[int] = None,
            chunksize: int = 1,
            use_threads: bool = False
            ) -> FrameOrSeries:
        '''
        {doc} Employ parallel processing with either the ProcessPoolExecutor or ThreadPoolExecutor.

        Args:
            {func}
            *
            {dtype}
            {name}
            {max_workers}
            {chunksize}
            {use_threads}
        '''
        # only use when we need pairs of values to dynamically create an Index
        if IterNodeApplyType.is_items(self._apply_type):
            apply_func = self._apply_iter_items_parallel
        else:
            apply_func = self._apply_iter_parallel
        return self._apply_constructor(
                apply_func(func,
                        max_workers=max_workers,
                        chunksize=chunksize,
                        use_threads=use_threads,
                        ),
                dtype=dtype,
                name=name,
                index_constructor=index_constructor,
                )

    def __iter__(self) -> tp.Union[
            tp.Iterator[tp.Any],
            tp.Iterator[tp.Tuple[tp.Any, tp.Any]]
            ]:
        '''
        Return a generator based on the yield type.
        '''
        if self._yield_type is IterNodeType.VALUES:
            yield from self._func_values()
        else:
            yield from self._func_items()


class IterNodeDelegateMapable(IterNodeDelegate[FrameOrSeries]):
    '''
    Delegate returned from :obj:`static_frame.IterNode`, providing iteration as well as a family of apply methods.
    '''

    __slots__ = ()

    INTERFACE = IterNodeDelegate.INTERFACE + (
            'map_all',
            'map_all_iter',
            'map_all_iter_items',
            'map_any',
            'map_any_iter',
            'map_any_iter_items',
            'map_fill',
            'map_fill_iter',
            'map_fill_iter_items',
            )

    @doc_inject(selector='map_any')
    def map_any_iter_items(self,
            mapping: Mapping
            ) -> tp.Iterator[tp.Tuple[tp.Any, tp.Any]]:
        '''
        {doc} A generator of resulting key, value pairs.

        Args:
            {mapping}
        '''
        get = getattr(mapping, 'get')
        if self._yield_type is IterNodeType.VALUES:
            yield from ((k, get(v, v)) for k, v in self._func_items())
        else:
            yield from ((k, get((k,  v), v)) for k, v in self._func_items())

    @doc_inject(selector='map_any')
    def map_any_iter(self,
            mapping: Mapping,
            ) -> tp.Iterator[tp.Any]:
        '''
        {doc} A generator of resulting values.

        Args:
            {mapping}
        '''
        get = getattr(mapping, 'get')
        if self._yield_type is IterNodeType.VALUES:
            yield from (get(v, v) for v in self._func_values())
        else:
            yield from (get((k,  v), v) for k, v in self._func_items())

    @doc_inject(selector='map_any')
    def map_any(self,
            mapping: Mapping,
            *,
            dtype: DtypeSpecifier = None,
            name: NameType = None,
            index_constructor: tp.Optional[IndexConstructor] = None,
            ) -> FrameOrSeries:
        '''
        {doc} Returns a new container.

        Args:
            {mapping}
            {dtype}
        '''
        if IterNodeApplyType.is_items(self._apply_type):
            return self._apply_constructor(
                    self.map_any_iter_items(mapping),
                    dtype=dtype,
                    index_constructor=index_constructor,
                    name=name,
                    )
        return self._apply_constructor(
                self.map_any_iter(mapping),
                dtype=dtype,
                index_constructor=index_constructor,
                name=name,
                )

    #---------------------------------------------------------------------------
    @doc_inject(selector='map_fill')
    def map_fill_iter_items(self,
            mapping: Mapping,
            *,
            fill_value: tp.Any = np.nan,
            ) -> tp.Iterator[tp.Tuple[tp.Any, tp.Any]]:
        '''
        {doc} A generator of resulting key, value pairs.

        Args:
            {mapping}
            {fill_value}
        '''
        get = getattr(mapping, 'get')
        if self._yield_type is IterNodeType.VALUES:
            yield from ((k, get(v, fill_value)) for k, v in self._func_items())
        else:
            yield from ((k, get((k,  v), fill_value)) for k, v in self._func_items())

    @doc_inject(selector='map_fill')
    def map_fill_iter(self,
            mapping: Mapping,
            *,
            fill_value: tp.Any = np.nan,
            ) -> tp.Iterator[tp.Any]:
        '''
        {doc} A generator of resulting values.

        Args:
            {mapping}
            {fill_value}
        '''
        get = getattr(mapping, 'get')
        if self._yield_type is IterNodeType.VALUES:
            yield from (get(v, fill_value) for v in self._func_values())
        else:
            yield from (get((k,  v), fill_value) for k, v in self._func_items())

    @doc_inject(selector='map_fill')
    def map_fill(self,
            mapping: Mapping,
            *,
            fill_value: tp.Any = np.nan,
            dtype: DtypeSpecifier = None,
            name: NameType = None,
            index_constructor: tp.Optional[IndexConstructor] = None,
            ) -> FrameOrSeries:
        '''
        {doc} Returns a new container.

        Args:
            {mapping}
            {fill_value}
            {dtype}
        '''
        if IterNodeApplyType.is_items(self._apply_type):
            return self._apply_constructor(
                    self.map_fill_iter_items(mapping, fill_value=fill_value),
                    dtype=dtype,
                    name=name,
                    index_constructor=index_constructor,
                    )
        return self._apply_constructor(
                self.map_fill_iter(mapping, fill_value=fill_value),
                dtype=dtype,
                name=name,
                index_constructor=index_constructor,
                )


    #---------------------------------------------------------------------------
    @doc_inject(selector='map_all')
    def map_all_iter_items(self,
            mapping: Mapping
            ) -> tp.Iterator[tp.Tuple[tp.Any, tp.Any]]:
        '''
        {doc} A generator of resulting key, value pairs.

        Args:
            {mapping}
        '''
        # want exception to raise if key not found
        func = getattr(mapping, '__getitem__')
        if self._yield_type is IterNodeType.VALUES:
            yield from ((k, func(v)) for k, v in self._func_items())
        else:
            yield from ((k, func((k,  v))) for k, v in self._func_items())

    @doc_inject(selector='map_all')
    def map_all_iter(self,
            mapping: Mapping
            ) -> tp.Iterator[tp.Any]:
        '''
        {doc} A generator of resulting values.

        Args:
            {mapping}
        '''
        func = getattr(mapping, '__getitem__')
        if self._yield_type is IterNodeType.VALUES:
            yield from (func(v) for v in self._func_values())
        else:
            yield from (func((k,  v)) for k, v in self._func_items())

    @doc_inject(selector='map_all')
    def map_all(self,
            mapping: Mapping,
            *,
            dtype: DtypeSpecifier = None,
            name: NameType = None,
            index_constructor: tp.Optional[IndexConstructor] = None,
            ) -> FrameOrSeries:
        '''
        {doc} Returns a new container.

        Args:
            {mapping}
            {dtype}
        '''
        if IterNodeApplyType.is_items(self._apply_type):
            return self._apply_constructor(
                    self.map_all_iter_items(mapping),
                    dtype=dtype,
                    name=name,
                    index_constructor=index_constructor,
                    )
        return self._apply_constructor(
                self.map_all_iter(mapping),
                dtype=dtype,
                name=name,
                index_constructor=index_constructor,
                )

#-------------------------------------------------------------------------------

class IterNode(tp.Generic[FrameOrSeries]):
    '''Interface to a type of iteration on :obj:`static_frame.Series` and :obj:`static_frame.Frame`.
    '''
    # Stores two version of a generator function: one to yield single values, another to yield items pairs. The latter is needed in all cases, as when we use apply we return a Series, and need to have recourse to an index.

    __slots__ = (
        '_container',
        '_func_values',
        '_func_items',
        '_yield_type',
        '_apply_type',
        )
    CLS_DELEGATE = IterNodeDelegate

    def __init__(self, *,
            container: FrameOrSeries,
            function_values: tp.Callable[..., tp.Iterable[tp.Any]],
            function_items: tp.Callable[..., tp.Iterable[tp.Tuple[tp.Any, tp.Any]]],
            yield_type: IterNodeType,
            apply_type: IterNodeApplyType,
            ) -> None:
        '''
        Args:
            function_values: will be partialed with arguments given with __call__.
            function_items: will be partialed with arguments given with __call__.
        '''
        self._container: FrameOrSeries = container
        self._func_values = function_values
        self._func_items = function_items
        self._yield_type = yield_type
        self._apply_type = apply_type

    #---------------------------------------------------------------------------
    # apply constructors

    def to_series_from_values(self,
            values: tp.Iterator[tp.Any],
            *,
            dtype: DtypeSpecifier,
            name: NameType = None,
            index_constructor: tp.Optional[IndexConstructor] = None,
            axis: int = 0,
            ) -> 'Series':
        from static_frame.core.series import Series

        # Creating a Series that will have the same index as source container
        if self._container._NDIM == 2 and axis == 0:
            index = self._container._columns #type: ignore
            own_index = False
        else:
            index = self._container._index
            own_index = True

        if index_constructor is not None:
            index = index_constructor(index)

        # PERF: passing count here permits faster generator realization
        values, _ = iterable_to_array_1d(
                values,
                count=index.shape[0],
                dtype=dtype,
                )
        return Series(values,
                name=name,
                index=index,
                own_index=own_index,
                )

    def to_series_from_items(self,
            pairs: tp.Iterable[tp.Tuple[tp.Hashable, tp.Any]],
            *,
            dtype: DtypeSpecifier = None,
            name: NameType = None,
            index_constructor: tp.Optional[IndexConstructor]= None,
            axis: int = 0,
            ) -> 'Series':
        from static_frame.core.series import Series

        # apply_constructor should be implemented to take a pairs of label, value; only used for iter_window
        # axis 0 iters windows labelled by the index, axis 1 iters windows labelled by the columns

        if self._container._NDIM == 2 and axis == 1:
            index_constructor = (index_constructor
                    if index_constructor is not None
                    else self._container._columns.from_labels) #type: ignore
            name_index = self._container._columns._name #type: ignore
        else:
            index_constructor = (index_constructor
                    if index_constructor is not None
                    else self._container._index.from_labels)
            name_index = self._container._index._name

        index_constructor_final = partial(
                index_constructor,
                name=name_index,
                )
        # always return a Series
        return Series.from_items(
                pairs=pairs,
                dtype=dtype,
                name=name,
                index_constructor=index_constructor_final,
                )

    def to_series_from_group_items(self,
            pairs: tp.Iterable[tp.Tuple[tp.Hashable, tp.Any]],
            *,
            dtype: DtypeSpecifier = None,
            name: NameType = None,
            index_constructor: tp.Optional[IndexConstructor]= None,
            name_index: NameType = None,
            ) -> 'Series':
        from static_frame.core.index import Index
        from static_frame.core.series import Series

        # NOTE: when used on labels, this key is given; when used on labels (indices) depth_level is given; only take the key if it is a hashable (a string or a tuple, not a slice, list, or array)

        index_constructor = partial(
                Index if index_constructor is None else index_constructor,
                name=name_index,
                )
        return Series.from_items(
                pairs=pairs,
                dtype=dtype,
                name=name,
                index_constructor=index_constructor
                )

    def to_frame_from_elements(self,
            items: tp.Iterable[tp.Tuple[
                    tp.Tuple[tp.Hashable, tp.Hashable], tp.Any]],
            *,
            dtype: DtypeSpecifier = None,
            name: NameType = None,
            index_constructor: tp.Optional[IndexConstructor]= None,
            axis: int = 0,
            ) -> 'Frame':
        from static_frame.core.frame import Frame

        index_constructor = (self._container._index.from_labels
                if index_constructor is None else index_constructor)

        assert isinstance(self._container, Frame)
        return self._container.__class__.from_element_items(
                items,
                index=self._container._index,
                columns=self._container._columns,
                axis=axis,
                own_index=True,
                index_constructor=index_constructor,
                columns_constructor=self._container._columns.from_labels,
                name=name,
                )

    def to_index_from_labels(self,
            values: tp.Iterator[tp.Hashable], #pylint: disable=function-redefined
            dtype: DtypeSpecifier = None,
            name: NameType = None,
            index_constructor: tp.Optional[IndexConstructor]= None,
            ) -> np.ndarray:
        # NOTE: name argument is for common interface
        if index_constructor is not None:
            raise RuntimeError('index_constructor not supported with this interface')
        # PERF: passing count here permits faster generator realization
        shape = self._container.shape
        array, _ = iterable_to_array_1d(values, count=shape[0], dtype=dtype)
        return array

    #---------------------------------------------------------------------------
    def _get_delegate_kwargs(self,
            **kwargs: object,
            ) -> tp.Dict[str, tp.Any]:
        '''
        In usage as an iteator, the args passed here are expected to be argument for the core iterators, i.e., axis arguments.

        Args:
            kwargs: kwarg args to be passed to both self._func_values and self._func_items
        '''
        from static_frame.core.frame import Frame

        # all kwargs, like ``drop```, are partialed into func_values, func_items
        func_values = partial(self._func_values, **kwargs)
        func_items = partial(self._func_items, **kwargs)

        axis = kwargs.get('axis', 0)

        apply_constructor: tp.Callable[..., tp.Union['Frame', 'Series']]

        if self._apply_type is IterNodeApplyType.SERIES_VALUES:
            apply_constructor = partial(self.to_series_from_values, axis=axis)

        elif self._apply_type is IterNodeApplyType.SERIES_ITEMS:
            apply_constructor = partial(self.to_series_from_items, axis=axis)

        elif self._apply_type is IterNodeApplyType.SERIES_ITEMS_GROUP_VALUES:
            try:
                name_index = name_filter(kwargs.get('key', None))
            except TypeError:
                name_index = None
            apply_constructor = partial(self.to_series_from_group_items,
                    name_index=name_index,
                    )

        elif self._apply_type is IterNodeApplyType.SERIES_ITEMS_GROUP_LABELS:
            # will always have `depth_level` in kwargs, and for Frame an axis; could attempt to get name from the index if it has a name
            name_index = None
            apply_constructor = partial(self.to_series_from_group_items,
                    name_index=name_index,
                    )

        elif self._apply_type is IterNodeApplyType.FRAME_ELEMENTS:
            assert isinstance(self._container, Frame) # for typing
            apply_constructor = partial(self.to_frame_from_elements, axis=axis)

        elif self._apply_type is IterNodeApplyType.INDEX_LABELS:
            apply_constructor = self.to_index_from_labels

        else:
            raise NotImplementedError(self._apply_type) #pragma: no cover

        return dict(
                func_values=func_values,
                func_items=func_items,
                yield_type=self._yield_type,
                apply_constructor=tp.cast(tp.Callable[..., FrameOrSeries], apply_constructor),
                apply_type=self._apply_type,
                )

    def get_delegate(self,
            **kwargs: object,
            ) -> IterNodeDelegate[FrameOrSeries]:
        return IterNodeDelegate(**self._get_delegate_kwargs(**kwargs))

    def get_delegate_mapable(self,
            **kwargs: object,
            ) -> IterNodeDelegateMapable[FrameOrSeries]:
        return IterNodeDelegateMapable(**self._get_delegate_kwargs(**kwargs))

#-------------------------------------------------------------------------------
# specialize IterNode based on arguments given to __call__

class IterNodeNoArg(IterNode[FrameOrSeries]):

    __slots__ = ()
    CLS_DELEGATE = IterNodeDelegate

    def __call__(self,
            ) -> IterNodeDelegate[FrameOrSeries]:
        return IterNode.get_delegate(self)


class IterNodeNoArgMapable(IterNode[FrameOrSeries]):

    __slots__ = ()
    CLS_DELEGATE = IterNodeDelegateMapable

    def __call__(self,
            ) -> IterNodeDelegateMapable[FrameOrSeries]:
        return IterNode.get_delegate_mapable(self)

class IterNodeAxisElement(IterNode[FrameOrSeries]):

    __slots__ = ()
    CLS_DELEGATE = IterNodeDelegateMapable

    def __call__(self,
            *,
            axis: int = 0
            ) -> IterNodeDelegateMapable[FrameOrSeries]:
        return IterNode.get_delegate_mapable(self, axis=axis)

class IterNodeAxis(IterNode[FrameOrSeries]):

    __slots__ = ()

    def __call__(self,
            *,
            axis: int = 0
            ) -> IterNodeDelegateMapable[FrameOrSeries]:
        return IterNode.get_delegate_mapable(self, axis=axis)

class IterNodeConstructorAxis(IterNode[FrameOrSeries]):

    __slots__ = ()
    CLS_DELEGATE = IterNodeDelegateMapable

    def __call__(self,
            *,
            axis: int = 0,
            constructor: tp.Optional[TupleConstructorType] = None,
            ) -> IterNodeDelegateMapable[FrameOrSeries]:
        return IterNode.get_delegate_mapable(self,
                axis=axis,
                constructor=constructor,
                )

class IterNodeGroup(IterNode[FrameOrSeries]):
    '''
    Iterator on 1D groupings where no args are required (but axis is retained for compatibility)
    '''

    __slots__ = ()

    def __call__(self,
            *,
            axis: int = 0
            ) -> IterNodeDelegate[FrameOrSeries]:
        return IterNode.get_delegate(self, axis=axis)

class IterNodeGroupAxis(IterNode[FrameOrSeries]):
    '''
    Iterator on 2D groupings where key and axis are required.
    '''

    __slots__ = ()

    def __call__(self,
            key: KEY_ITERABLE_TYPES, # type: ignore
            *,
            axis: int = 0,
            drop: bool = False,
            ) -> IterNodeDelegate[FrameOrSeries]:
        return IterNode.get_delegate(self, key=key, axis=axis, drop=drop)


class IterNodeGroupOther(IterNode[FrameOrSeries]):
    '''
    Iterator on 1D groupings where group values are provided.
    '''

    __slots__ = ()

    def __call__(self,
            other: tp.Union[np.ndarray, 'Index', 'Series', tp.Iterable[tp.Any]],
            *,
            fill_value: tp.Any = np.nan,
            axis: int = 0
            ) -> IterNodeDelegate[FrameOrSeries]:

        index_ref = (self._container._index if axis == 0
                else self._container._columns) # type: ignore

        group_source = group_from_container(
                index=index_ref,
                group_source=other,
                fill_value=fill_value,
                axis=axis,
                )

        # kwargs are partialed into finc_values, func_items
        return IterNode.get_delegate(self,
                axis=axis,
                group_source=group_source,
                )

class IterNodeDepthLevel(IterNode[FrameOrSeries]):

    __slots__ = ()

    def __call__(self,
            depth_level: tp.Optional[DepthLevelSpecifier] = None
            ) -> IterNodeDelegateMapable[FrameOrSeries]:
        return IterNode.get_delegate_mapable(self, depth_level=depth_level)


class IterNodeDepthLevelAxis(IterNode[FrameOrSeries]):

    __slots__ = ()

    def __call__(self,
            depth_level: DepthLevelSpecifier = 0,
            *,
            axis: int = 0
            ) -> IterNodeDelegate[FrameOrSeries]:
        return IterNode.get_delegate(self, depth_level=depth_level, axis=axis)


class IterNodeWindow(IterNode[FrameOrSeries]):

    __slots__ = ()

    def __call__(self, *,
            size: int,
            axis: int = 0,
            step: int = 1,
            window_sized: bool = True,
            window_func: tp.Optional[AnyCallable] = None,
            window_valid: tp.Optional[AnyCallable] = None,
            label_shift: int = 0,
            start_shift: int = 0,
            size_increment: int = 0,
            ) -> IterNodeDelegate[FrameOrSeries]:
        return IterNode.get_delegate(self,
                axis=axis,
                size=size,
                step=step,
                window_sized=window_sized,
                window_func=window_func,
                window_valid=window_valid,
                label_shift=label_shift,
                start_shift=start_shift,
                size_increment=size_increment,
                )

